/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Master of Science - Computergraphics and Image Processing
 *
 * Robotics - Prof. Dr. Max Fischer
 * "Autonomous self-parking robot"
 * 
 * System: Linux, C++
 */

/**
 * USABILITY:
 *  Copyright (c) 2006 
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * Fibonacci Heap Structure, to be used in the AStarPlaner as support
 * for the slow openList map implementation. Uses the Fibonacci Heap implementation
 * of Daniel Berlin (dan@cgsoftware.com) programmed for GCC "libiberty". Needs -liberty
 * as linker flag.
 *
 *	Information by Daniel Berlin (taken from fibheap.h)
 *    	Amortized and real worst case time for operations:
 *
 * 	ExtractMin: O(lg n) amortized. O(n) worst case.
 *  	DecreaseKey: O(1) amortized.  O(lg n) worst case. 
 *  	Insert: O(2) amortized. O(1) actual.  
 *  	Union: O(1) amortized. O(1) actual.
 *
 * @author Thomas Wöllert, Beatrix Müller
 */

#ifndef ASTAR_FIBHEAP_H
#define ASTAR_FIBHEAP_H

#include <iostream>

extern "C" {
	#include "fibheap.h"
}

#include "AStarNode.h"

using namespace std;

class AStarFibHeap {

	protected: 
		// The fibonacci heap to use
		fibheap_t heap;

	public:
		/**
		* Constructor
		*/
		AStarFibHeap();
		
		/**
		* Destructor
		*/
		virtual ~AStarFibHeap();

		/**
		* Insert a new node with the given heuristic into the heap
		* @param AStarNode* the node to insert
		* @param long the nodes heuristic by which the heap is sorted
		*/
		void insert(AStarNode* _node, long _heuristic);
		
		/**
		* Remove a node from the heap
		* @param AStarNode* the node to remove
		*/
		void erase(AStarNode* _node);
		
		/**
		* Checks if the heap is currently empty
		* @return true if it is empty, false otherwise
		*/
		bool empty();
		
		/**
		* Removes all elements from the heap
		*/
		void clear();
		
		/**
		* Tell the heap to re-sort a given node because
		* the nodes heuristic changed.
		* @param AStarNode* the node which heuristic changed
		* @param long the changed heuristic
		*/
		void setHeuristic(AStarNode* _node, long _heuristic);
		
		/**
		* Retrieve the node with the currently min-heuristic
		* (also removes the node from the heap).
		* @return AStarNode* the current min-heuristic node
		*/
		AStarNode* getBestHeuristicNode();

};

#endif // ASTAR_FIBHEAP_H
