/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Master of Science - Computergraphics and Image Processing
 *
 * Robotics - Prof. Dr. Max Fischer
 * "Autonomous self-parking robot"
 * 
 * System: Linux, C++
 */

/**
 * USABILITY:
 *  Copyright (c) 2006 
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef ASTARNODE_H
#define ASTARNODE_H

/**
 * 'AStarNode' defines the nodes for the A* (A-Star)
 * algorithm providing information on its own ID, its
 * parent as well as on its G, H and F costs.
 *
 * @author Beatrix Müller, Thomas Wöllert
 */

#include <iostream>
#include "fibheap.h"

using namespace std;


class AStarNode {

	private:
		// The node id
		int m_nodeId;
		
		// Pointer to the parent node
		AStarNode* m_parent;
	
		// G-costs (moving costs)
		int m_g;
		
		// H-costs (estimated costs)
		int m_h;
		
	public:
		// Fibonacci heap node
		fibnode_t fibHeapNode;

	public:	
		/**
		* Default Constructor
		*/
		AStarNode() {}
		
		/**
		* Custom Constructor
		* @param int the node id
		* @param AStarNode* pointer to the parent node
		* @param int g-cost for this node
		* @param int h-cost for this node
		*/
		AStarNode(int _nodeId, AStarNode* _parent, int _g, int _h) {
			m_nodeId = _nodeId;
			m_parent = _parent;
			m_g	= _g;
			m_h	= _h;
		}
		
		/**
		* Copy Constructor
		* @param AStarNode& the node to copy
		*/
		AStarNode(const AStarNode& _rhs) {
			m_nodeId = _rhs.getNodeId();
			m_parent = _rhs.getParent();
			m_g = _rhs.getG();
			m_h = _rhs.getH();
		}
		
		/**
		* Destructor
		* (is not deleting the parent node)
		*/
		virtual ~AStarNode() {}

		/**
		* Overloading the operator=
		* @param AStarNode& the node to assign
		*/
		bool operator=(const AStarNode& _rhs) {
			m_nodeId = _rhs.getNodeId();
			m_parent = _rhs.getParent();
			m_g = _rhs.getG();
			m_h = _rhs.getH();
			return true;
		}
		
		/**
		 * Returns the node ID of the AStarNode
		 * @return the node ID.
		 */
		int getNodeId() const {
			return m_nodeId;
		}
		
		/**
		 * Returns the parent of the AStarNode
		 * @return the parent.
		 */
		AStarNode* getParent() const {	
			return m_parent;
		}
		
		/**
		 * Sets the parent of the AStarNode.
		 * @param AStarNode* the parent to set
		 */
		void setParent(AStarNode* _parent) {
			m_parent = _parent;
		}
		
		
		/**
		 * Sets the G cost of the AStarNode.
		 * @param int the G to set
		 */
		void setG(int _g) {
			m_g = _g;
		}
		
		/**
		 * Sets the H cost of the AStarNode.
		 * @param int the H to set
		 */
		void setH(int _h) {
			m_h = _h;
		}
		
		/**
		 * Returns the F cost of the AStarNode,
		 * which is the sum of G and H (F = G + H).
		 * @return the F cost.
		 */
		int getF() const {
			return (m_g + m_h);
		}
		
		/**
		 * Returns the G cost of the AStarNode.
		 * @return the G cost.
		 */
		int getG() const {
			return m_g;
		}

		/**
		 * Returns the H cost of the AStarNode.
		 * @return the H cost.
		 */
		int getH() const {
			return m_h;
		}
};


#endif
