#include "AStarPlaner.h"
#include "Car.h"
#include "Environment.h"
#include "CVCoordDrawingArea.h"
#include "Robot.h"
#include "Config.h"
#include <time.h>

extern Environment* g_Environment;
extern CVCoordDrawingArea* g_CoordDrawingArea;
extern Robot* g_Robot;

/*
*-------------------------------------------------------------------------------------------------------
*/
AStarPlaner::AStarPlaner(Car* _car, double _deltaT) {
	m_car = _car;
	m_deltaT = _deltaT;
	m_foundPath = new vector<Position*>();
	m_foundPath_rev = NULL;
	m_nodeCounter = 0;
}

AStarPlaner::~AStarPlaner() {

	if(m_foundPath != NULL) {
		cout << "Deleting found path ..." << endl;
		m_foundPath->clear();
		delete m_foundPath;
	}

	if (m_foundPath_rev != NULL) {
		cout << "Deleteing reverse path ..." << endl;
		// No need to clear this list, already cleared in m_foundPath
		delete m_foundPath_rev;
	}
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void AStarPlaner::run() {

	cout << "Initializing AStarPlaner ..." << endl; 
	
	// Area properties for both search runs
	double width = g_Environment->getAreaWidth();
	double height = g_Environment->getAreaHeight();
	double angle = g_Environment->getAreaAngle();
	
	//
	// DRAW PARKING LINES
	//
	
	Point* parkingLines = g_Environment->getParkingLines();
	
	if(parkingLines == NULL) {
		cout << "No parking lot found ..." << endl;
		return;
	}	
	
	g_CoordDrawingArea->drawParkingLines(parkingLines, 6, 1, g_CoordDrawingArea->createColor("black"));
	
	//
	// DO GRID SEARCH
	//
	
	// Start and Target for the first run
	Position startPosition = g_Environment->getStartPosition(true);
	Position targetPosition = g_Environment->getTargetPosition();
	
	// Draw the points
	g_CoordDrawingArea->drawPoint(startPosition.getX(), startPosition.getY(), 3, g_CoordDrawingArea->createColor("red"));
	g_CoordDrawingArea->drawPoint(targetPosition.getX(), targetPosition.getY(), 3, g_CoordDrawingArea->createColor("green"));
	g_CoordDrawingArea->updateWindow();
	
	// Create the first area
	m_area = new Area(m_car, m_deltaT, width, height, angle, GRID_WIDTH_STEPS, GRID_HEIGHT_STEPS, GRID_ANGLE_STEPS);

	// Initiate the A*-algorithm
	clock_t startTime = clock();
	Position* reachedPos = findPath(startPosition, targetPosition, COST_LEVEL_CHANGE, COST_DIRECTION_CHANGE, TARGET_X_THRESHOLD, TARGET_Y_THRESHOLD, TARGET_THETA_THRESHOLD);
	
	if(reachedPos == NULL) {
		cout << "### Error during A*-Search ###" << endl;
		return;
	}
	
	clock_t endTime = clock();
	
	double neededTime = ((double)(endTime - startTime)) / CLOCKS_PER_SEC;
	
	cout << "### A*-Search complete ###" << endl;
	cout << "Calculation time: " << neededTime << " sec" << endl;
	cout << "Nodes processed: " << m_nodeCounter << endl;
	
	if(neededTime > 0) {
		cout << "Nodes per second: " << (((double)m_nodeCounter)/neededTime) << " nodes/sec" << endl;
	}
	cout << "Reached position (x/y/theta): " << reachedPos->getX() << " / " << reachedPos->getY() << " / " << reachedPos->getTheta() << endl;
	cout << "######" << endl;
	
	if(reachedPos != NULL) {
		delete reachedPos;
	}
	
	if(m_area != NULL) {
		delete m_area;
	}
		
	//
	// DISPLAY PATH
	//
	
	// Display the reslts
	g_CoordDrawingArea->clear();
	g_CoordDrawingArea->drawGrid(1.0, 1.0, 1, g_CoordDrawingArea->createColor("lightgray"));
	g_CoordDrawingArea->drawGridText(1.0, 1.0, g_CoordDrawingArea->createColor("lightgray"));
	g_CoordDrawingArea->updateWindow();
	
	g_CoordDrawingArea->drawParkingLines(g_Environment->getParkingLines(), 6, 1, g_CoordDrawingArea->createColor("black"));
	
	startPosition = g_Environment->getStartPosition(false);
	targetPosition = g_Environment->getTargetPosition();
	
	g_CoordDrawingArea->drawPoint(startPosition.getX(), startPosition.getY(), 3, g_CoordDrawingArea->createColor("red"));
	g_CoordDrawingArea->drawPoint(targetPosition.getX(), targetPosition.getY(), 3, g_CoordDrawingArea->createColor("green"));
	
	Position startPos = *(m_foundPath->at(0));
	
	for(unsigned int i = 1; i < m_foundPath->size(); i++) {
		Position endPos = *(m_foundPath->at(i));
		
		g_CoordDrawingArea->drawLine(startPos.getX(), startPos.getY(), endPos.getX(), endPos.getY(), 1, g_CoordDrawingArea->createColor("black"));
		startPos = endPos;
	}
	
	//
	// DRIVE PATH
	//
	
	cout << "Driving Path ..." << endl;
	m_car->drivePath(m_deltaT, m_foundPath);
	Position currentPos = g_Robot->getCurrentPosition();
	cout << "Path complete - Reached position (x/y/theta): " << currentPos.getX() << " / " << currentPos.getY() << " / " << currentPos.getTheta() << endl;
	cout << "######" << endl;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
Position* AStarPlaner::findPath(Position _startPosition, Position _targetPosition, int _costLevelChange, int _costDirectionChange, double _xThreshold, double _yThreshold, double _thetaThreshold) {

	cout << "Starting Pathfinding ..." << endl;
	m_nodeCounter = 0;

	m_costLevelChange = _costLevelChange;
	m_costDirectionChange = _costDirectionChange;
	
	//	
	// OFFSET CALCULATION
	//
	
	// Calculate the start position in the area
	m_area->clearGridOffset();
	AreaPosition startOffsetPos = m_area->mapPositionToArea(&_startPosition);
	m_area->setGridOffset(startOffsetPos.row, startOffsetPos.col, startOffsetPos.level);

	//
	// START THE REAL PLANER
	//
	
	// Map the positions into the area
	AreaPosition startAreaPos = m_area->mapPositionToArea(&_startPosition);
	AreaPosition targetAreaPos = m_area->mapPositionToArea(&_targetPosition);
	
	#ifdef DEBUG
		printf("%s %d: ROWS    : %d\n", __FILE__, __LINE__, area->getRows());
		printf("%s %d: COLUMNS : %d\n", __FILE__, __LINE__, area->getColumns());
		printf("%s %d: LEVELS  : %d\n", __FILE__, __LINE__, area->getLevels());
	#endif

	int startNodeId = m_area->getNodeId(startAreaPos.row, startAreaPos.col, startAreaPos.level);
	int targetNodeId = m_area->getNodeId(targetAreaPos.row, targetAreaPos.col, targetAreaPos.level);
	
	// Make sure that the start is within the grid
	if(startNodeId == -1) {
		cout << "Failure: Start node is not within grid scene: " << startNodeId << endl;
		return NULL;
	}
	
	// Make sure that the target is within the grid
	if(targetNodeId == -1) {
		cout << "Failure: Target node is not within grid scene: " << targetNodeId << endl;
		return NULL;
	}
	
	// generate the start and target node
	AreaNode* areaStartNode = new AreaNode(startNodeId, *(startAreaPos.pos), false);
	AreaNode* areaTargetNode = new AreaNode(targetNodeId, *(targetAreaPos.pos), false);
	
	m_area->setStartNode(areaStartNode);
	m_area->setTargetNode(areaTargetNode);
	
	int h = getHeuristic(m_area->getStartNodeId());
	AStarNode* startNode = new AStarNode(startNodeId, NULL, 0, h);	
	
	// ----------------------------------------------
	
	// OpenList and ClosedList
	AStarNodeMap openList;	
	AStarNodeMap closedList;

	// Add start node to OpenList
	openList.insert(AStarNodeMapPair(startNode->getNodeId(), startNode));
	#ifdef USE_FIBONACCI_HEAP
		fibHeap.insert(startNode, startNode->getF());
	#endif
	
	// Remember the starting time
	clock_t startTime = clock();
	
	// until target node not found:
	while (!openList.empty()) {
	
		// search in open list for node with lowest f
		AStarNode* actualNode = getBestNodeFromOpenList(&openList);
		
		if (actualNode == NULL) {
			// openlist is empty, no more actual nodes available
			fprintf(stderr, "%s %d: The OpenList is empty - no more nodes available. No target found.\n", __FILE__, __LINE__);
			
			// Mob up
			clearMap(&openList);
			clearMap(&closedList);
			#ifdef USE_FIBONACCI_HEAP
				fibHeap.clear();
			#endif
			
			return NULL;
		}
		
		int actualNodeId = actualNode->getNodeId();
		AreaNode* currentAreaNode = m_area->getNode(actualNodeId);
		Position currentPosition = currentAreaNode->getPosition();
		
		#ifdef DRAW
			g_CoordDrawingArea->drawPoint(currentPosition.getX(), currentPosition.getY(), 3, g_CoordDrawingArea->createColor("black"));
			g_CoordDrawingArea->updateWindow();
		#endif
		
		// New node to process, raise the node counter
		m_nodeCounter++;
	
		// Move this node to the ClosedList
		closedList.insert(AStarNodeMapPair(actualNode->getNodeId(), actualNode));
		
		// Remove this node from the OpenList
		openList.erase(actualNodeId);
		#ifdef USE_FIBONACCI_HEAP
			fibHeap.erase(actualNode);
		#endif
		
		// Test if actual node ID is within target area threshold
		AreaNode* targetAreaNode = m_area->getNode(m_area->getTargetNodeId());
		Position targetPosition = targetAreaNode->getPosition();
		
		if(currentPosition.getX() <= (targetPosition.getX()+_xThreshold) && currentPosition.getX() >= (targetPosition.getX()-_xThreshold)) {		
			if(currentPosition.getY() <= (targetPosition.getY()+_yThreshold) && currentPosition.getY() >= (targetPosition.getY()-_yThreshold)) {
				if(currentPosition.getTheta() <= (targetPosition.getTheta()+_thetaThreshold) && currentPosition.getTheta() >= (targetPosition.getTheta()-_thetaThreshold)) {
				
					cout << "Found path: Target within given threshold ..." << endl;;
				
					vector<int>* path = savePath(startNodeId, actualNode);
					savePositionPath(path);
					Position reachedPos = (m_area->getNode(actualNodeId)->getPosition());
					
					// Mob up
					clearMap(&openList);
					clearMap(&closedList);
					#ifdef USE_FIBONACCI_HEAP
						fibHeap.clear();
					#endif
					
					return new Position(reachedPos.getX(), reachedPos.getY(), reachedPos.getTheta());
				}		
			}	
		}
		
		// Check if target node is already in ClosedList
		int targetNodeId = m_area->getTargetNodeId();
		
		AStarNode* tmpNode = isInList(&closedList, targetNodeId);	
		if (tmpNode != NULL) {
		
			cout << "Found path: Target node is within the closed list (check #1) ..." << endl;
		
			// We found a path, save it
			vector<int>* path = savePath(startNodeId, tmpNode);
			savePositionPath(path);

			// Return the reached position
			Position reachedPos = m_area->getNode(tmpNode->getNodeId())->getPosition();
			
			// Mob up
			clearMap(&openList);
			clearMap(&closedList);
			#ifdef USE_FIBONACCI_HEAP
				fibHeap.clear();
			#endif
			
			return new Position(reachedPos.getX(), reachedPos.getY(), reachedPos.getTheta());
		}
		
		// For each adjacent node
		vector<AreaNode*> successors = vector<AreaNode*>();
		m_area->getSuccessors(actualNodeId, &successors);
		
		for (unsigned int i = 0; i < successors.size(); ++i) {
			
			AreaNode* areanode = successors.at(i);
			int areaNodeId = areanode->getNodeId();
			
			if (areaNodeId == -1) {
				fprintf(stderr, "%s %d: Invalid node ID of AreaNode.\n", __FILE__, __LINE__);
				
				// Mob up
				clearMap(&openList);
				clearMap(&closedList);
				#ifdef USE_FIBONACCI_HEAP
					fibHeap.clear();
				#endif
				
				return NULL;
			}
			
			// if not accessible or in ClosedList -> ignore
			if (areanode->isObstacle() || isInList(&closedList, areaNodeId) != NULL) {
				// ignore the node
			}
			else {
				
				// if not in open list -> add (node, parent, f, g, h)
				AStarNode* tmpNode = isInList(&openList, areaNodeId);
				if (tmpNode == NULL) {
					
					// the the h cost
					int h = getHeuristic(areaNodeId);
					
					// add the node with the information on id, parent, G and H to the OpenList
					AStarNode* node = new AStarNode(areaNodeId, actualNode, actualNode->getG()+getCosts(actualNode, actualNode->getParent()), h);
					openList.insert(AStarNodeMapPair(node->getNodeId(), node));
					#ifdef USE_FIBONACCI_HEAP
						fibHeap.insert(node, node->getF());
					#endif
				}
				else {

					AStarNode* node = isInList(&openList, areaNodeId);
					AStarNode* parent = node->getParent();
					
					// if true: add parent node to actual node and get new G and F
					if (actualNode->getG() < parent->getG()) {
						
						// set the actual node as parent
						node->setParent(actualNode);
						
						// set the new G cost
						node->setG(actualNode->getG()+getCosts(node, parent));
						
						// set the new H cost
						int h = getHeuristic(node->getNodeId());
						node->setH(h);
						
						#ifdef USE_FIBONACCI_HEAP
							fibHeap.setHeuristic(node, node->getG());
						#endif
					}
				}
			}
		}
		
		// Give the user some status information
		clock_t endTime = clock();
		
		if(m_nodeCounter%10000 == 0) {
			double neededTime = ((double)(endTime - startTime)) / CLOCKS_PER_SEC;
			cout << neededTime << " sec : " << m_nodeCounter << " Nodes processed at " << (m_nodeCounter/neededTime) << " Nodes/sec " << endl;
		}
		
		// finish process if:
		// target node moved to closed list
		// target node could not be found and open list is empty -> no path
		tmpNode = isInList(&closedList, targetNodeId);
		if (tmpNode != NULL) {
		
			cout << "Found path: Target node is within the closed list (check #2) ..." << endl;

			// Save the path
			vector<int>* path = savePath(startNodeId, tmpNode);
			savePositionPath(path);

			// Return the reached position
			Position reachedPos = m_area->getNode(tmpNode->getNodeId())->getPosition();
			
			// Mob up
			clearMap(&openList);
			clearMap(&closedList);
			#ifdef USE_FIBONACCI_HEAP
				fibHeap.clear();
			#endif
			
			return new Position(reachedPos.getX(), reachedPos.getY(), reachedPos.getTheta());
		} 
		else if (openList.empty()) {
			printf("%s %d: Target not reachable.\n", __FILE__, __LINE__);
			
			// Mob up
			clearMap(&openList);
			clearMap(&closedList);
			#ifdef USE_FIBONACCI_HEAP
				fibHeap.clear();
			#endif
			
			return NULL;
		}
	}
	printf("%s %d: The OpenList is empty - no more nodes available. No target found.\n", __FILE__, __LINE__);
	
	// Mob up
	clearMap(&openList);
	clearMap(&closedList);
	#ifdef USE_FIBONACCI_HEAP
		fibHeap.clear();
	#endif
	
	return NULL;
}

//
// ----------------------------------------------------------------------------------------------------------
//

vector<int>* AStarPlaner::savePath(const int startNodeId, const AStarNode* targetNode) {

	vector<int>* path = new vector<int>();
	
	printf("%s %d: Path found! :)\n", __FILE__, __LINE__);
	
	path->push_back(targetNode->getNodeId());
	AStarNode* parent = targetNode->getParent();
	
	while (parent->getNodeId() != startNodeId) {
		path->push_back(parent->getNodeId());
		parent = parent->getParent();
	}
	path->push_back(startNodeId);
	
	printf("%s %d: Path: ",  __FILE__, __LINE__);
	for (int i = (path->size()-1); i >= 0; i--) {
		printf("%d ", path->at(i));
	}
	printf("\n");
	
	return path;
}

//
// ----------------------------------------------------------------------------------------------------------
//

void AStarPlaner::savePositionPath(vector<int>* _nodeIdPath) {

	for(int i = (_nodeIdPath->size()-1); i >= 0; i--) {
		int currentNodeId = _nodeIdPath->at(i);
		
		AreaNode* currentAreaNode = m_area->getNode(currentNodeId);
		Position currentPos = currentAreaNode->getPosition();
		
		m_foundPath->push_back(new Position(currentPos.getX(), currentPos.getY(), currentPos.getTheta()));
	}
}

//
// ----------------------------------------------------------------------------------------------------------
//

int AStarPlaner::getHeuristic(const int nodeId) {

	int target = m_area->getTargetNodeId();
	
	int colStart = m_area->getColumn(nodeId);
	int colTarget = m_area->getColumn(target);
	int rowStart  = m_area->getRow(nodeId);
	int rowTarget = m_area->getRow(target);
	//int levelStart  = m_area->getLevel(nodeId);
	//int levelTarget = m_area->getLevel(target);
	
	int diffCol = abs(colTarget - colStart);
	int diffRow = abs(rowTarget - rowStart);
	
	// Diff level calculation has to be changed, theta range in real world: 0->2*PI
	// but 2*PI is the same as zero and does not have a big distance
	// in other words the area grid has a connection from theta level 0 to theta level max (i.e. 500)
	// and that only accounts for a distance of 1 instead of 500.
	
	//int diffLevel = abs(levelTarget - levelStart);
	
	// Calculate the angle difference
	AreaNode* startNode = m_area->getNode(nodeId);
	AreaNode* targetNode = m_area->getNode(target);
	
	double startTheta = startNode->getPosition().getTheta();
	double targetTheta = targetNode->getPosition().getTheta();
	
	double diffTheta = abs(targetTheta - startTheta);
	
	if(diffTheta >= M_PI) {
		diffTheta = abs(diffTheta - 2*M_PI);
	}
	
	double levelSteps = dtor(360.0) / m_area->getLevels();
	int diffLevel = diffTheta / levelSteps;
	
	// return Manhattan Distance
	return (diffCol + diffRow + diffLevel);
}

//
// ----------------------------------------------------------------------------------------------------------
//

AStarNode* AStarPlaner::getBestNodeFromOpenList(AStarNodeMap* nodeList) {

	#ifdef USE_FIBONACCI_HEAP
		if(!fibHeap.empty()) {
			return fibHeap.getBestHeuristicNode();
		}
	#else
		// get the node with the lowest F cost from the given list	
		if(!nodeList->empty()) {
		
			AStarNode* result = (*nodeList->begin()).second;
			int minF = result->getF();
		
			for(AStarNodeMap::iterator i = nodeList->begin(); i != nodeList->end(); ++i) {
				AStarNode* node = (*i).second;
				
				if(node->getF() < minF) {
					minF = node->getF();
					result = node;
				}
			}
			
			return result;
		}
	#endif
	
	return NULL;
}

//
// ----------------------------------------------------------------------------------------------------------
//

AStarNode* AStarPlaner::isInList(AStarNodeMap* nodeList, const int nodeId) {
	
	// Check if there is a node with the given ID in the given list	
	AStarNodeMap::iterator i = nodeList->find(nodeId);
	
	if(i != nodeList->end()) {
		return (*i).second;
	} else {
		return NULL;
	}
}

//
// ----------------------------------------------------------------------------------------------------------
//

void AStarPlaner::printList(AStarNodeMap* nodeList) {
	
	printf("[");
	
	for(AStarNodeMap::iterator i = nodeList->begin(); i != nodeList->end(); ++i) {

		AStarNode* node = (*i).second;
		int id = node->getNodeId();
		printf("%d ", id);
	}

	printf("]\n");
}

//
// ----------------------------------------------------------------------------------------------------------
//

int AStarPlaner::getCosts(AStarNode* node, AStarNode* parent) {

	int overallCosts = 10;

	if(parent == NULL) {
		return overallCosts;
	}
	
	// The current node
	int nodeId = node->getNodeId();
	AreaNode* areaNode = m_area->getNode(nodeId);
	int level = m_area->getLevel(nodeId);
	
	// The parent node
	int parentNodeId = parent->getNodeId();
	AreaNode* parentNode = m_area->getNode(parentNodeId);
	int parentLevel = m_area->getLevel(parentNodeId);
	
	// Determine costs for theta(level) change
	if(level != parentLevel) {
		int levelDiff = abs(level-parentLevel);
		overallCosts += levelDiff * m_costLevelChange;
	}
	
	if(parent->getParent() == NULL) {
		return overallCosts;
	}
	
	// The grandparent node
	AStarNode* grandParent = parent->getParent();
	int grandParentNodeId = grandParent->getNodeId();
	AreaNode* grandParentNode = m_area->getNode(grandParentNodeId);
	
	// Direction grandparent->parent
	Position rotGPPos = rotatePoint(parentNode->getPosition(), grandParentNode->getPosition(), -1.0*grandParentNode->getPosition().getTheta());
	
	int gPDriveDirection = 0;
	
	if(rotGPPos.getX() >= grandParentNode->getPosition().getX()) {
		// Forward
		gPDriveDirection = 1;
	} else {
		// Backwards
		gPDriveDirection = -1;
	}
	
	// Direction parent->current
	Position rotPCPos = rotatePoint(areaNode->getPosition(), parentNode->getPosition(), -1.0*parentNode->getPosition().getTheta());
	
	int pCDriveDirection = 0;
	
	if(rotPCPos.getX() >= parentNode->getPosition().getX()) {
		// Forward
		pCDriveDirection = 1;
	} else {
		// Backwards
		pCDriveDirection = -1;
	}

	if(gPDriveDirection != pCDriveDirection) {
		overallCosts += m_costDirectionChange;
	}
	
	return overallCosts;
}

//
// ----------------------------------------------------------------------------------------------------------
//

Position AStarPlaner::rotatePoint(Position _posToRotate, Position _rotateOrigin, double _angle) {

	_rotateOrigin.setX(-1.0*_rotateOrigin.getX());
	_rotateOrigin.setY(-1.0*_rotateOrigin.getY());

	Position transPoint;
	transPoint.setX(_posToRotate.getX() + _rotateOrigin.getX());
	transPoint.setY(_posToRotate.getY() + _rotateOrigin.getY());
	
	double newX = (transPoint.getX() * cos(_angle)) - (transPoint.getY() * sin(_angle));
	double newY = (transPoint.getX() * sin(_angle)) + (transPoint.getY() * cos(_angle));
	
	transPoint.setX(newX);
	transPoint.setY(newY);
	
	_rotateOrigin.setX(-1.0*_rotateOrigin.getX());
	_rotateOrigin.setY(-1.0*_rotateOrigin.getY());

	Position rotatedPoint = Position();
	rotatedPoint.setX(transPoint.getX() + _rotateOrigin.getX());
	rotatedPoint.setY(transPoint.getY() + _rotateOrigin.getY());
	rotatedPoint.setTheta(0.0);
	
	return rotatedPoint;
}

//
// ----------------------------------------------------------------------------------------------------------
//

void AStarPlaner::clearMap(AStarNodeMap* _map) {

	AStarNodeMap::iterator i = _map->begin();
	
	while(!_map->empty()) {
		AStarNode* node = (*i).second;
		_map->erase(i);
		
		if(node != NULL) {
			delete node;
		}
		
		i++;
	}
}

//
// ----------------------------------------------------------------------------------------------------------
//

void AStarPlaner::driveReverse() {

	
	//
	// DRIVE FROM PARKING LOT TO START
	//
	
	cout << "Drive reverse..." << endl;

	if (m_foundPath != NULL) {
	
		m_foundPath_rev = new vector<Position*>();
	
		int j = (m_foundPath->size() - 1);
		for (unsigned int i = 0; i < m_foundPath->size(); i++) {
			m_foundPath_rev->push_back(m_foundPath->at(j));
			--j;
		}
		
		cout << "Driving Path back to start point ..." << endl;
		m_car->drivePath(m_deltaT, m_foundPath_rev);
		Position currentPos = g_Robot->getCurrentPosition();
		cout << "Path complete - Reached position (x/y/theta): " << currentPos.getX() << " / " << currentPos.getY() << " / " << currentPos.getTheta() << endl;
		cout << "######" << endl;
	}
}
