/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Master of Science - Computergraphics and Image Processing
 *
 * Robotics - Prof. Dr. Max Fischer
 * "Autonomous self-parking robot"
 * 
 * System: Linux, C++
 */

/**
 * USABILITY:
 *  Copyright (c) 2006 
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * The main A*-planer which can be called
 *
 * @author Beatrix Müller, Thomas Wöllert
 */

#ifndef ASTARPLANER_H
#define ASTARPLANER_H

#include "AStarNode.h"
#include "Area.h"
#include "Car.h"
#include "AStarFibHeap.h"

#include <iostream>
#include <vector>
#include <map>

//#define DRAW
#define USE_FIBONACCI_HEAP

using namespace std;

// Definition of the AStarNode maps
typedef map<int, AStarNode*> AStarNodeMap;
typedef AStarNodeMap::value_type AStarNodeMapPair;

class AStarPlaner {

	protected:
		// The car model used in this planer
		Car* m_car;
		
		// The time step to use
		double m_deltaT;
		
		// The area
		Area* m_area;
		
		// The current costs for a level change
		int m_costLevelChange;
		
		// The current costs for a direction change
		int m_costDirectionChange;
		
		// Global position path
		vector<Position*>* m_foundPath;
		
		// Global position path reverse
		vector<Position*>* m_foundPath_rev;
		
		// Node counter (for the statistics)
		int m_nodeCounter;
		
		#ifdef USE_FIBONACCI_HEAP
			AStarFibHeap fibHeap;
		#endif
	
	public:
		/**
		* Constructor
		* @param Car* the car model to use
		* @param double the time step to use
		*/
		AStarPlaner(Car* _car, double _deltaT);
	
		/**
		* Destructor
		*/
		virtual ~AStarPlaner();
	
		/**
		* Start the algorithm
		*/
		virtual void run();
			
		/**
		* The A* (A-Star) pathfinder algorithm
		* The A* algorithm detects the shortest path from a given
		* start node to the given target node depending on costs
		* @param Position* the start position (in the real world)
		* @param Position* the target position (in the real world)
		* @param int costs of a level change (theta-change)
		* @param int costs of a direction change
		* @param double x-threshold for the target position
		* @param double y-threshold for the target position
		* @param double theta-threshold for the target position
		* @return the reached position.
		*/
		Position* findPath(Position _startPosition, Position _targetPosition, int _costLevelChange, int _costDirectionChange, double _xThreshold, double _yThresold, double _thetaThreshold);
		
		/**
		* Saves the found path as a vector of node IDs
		* @param int the start node ID
		* @param AStarNode* the target node
		* @return the found path (the nodes as a vector).
		*/
		vector<int>* savePath(const int _startNodeId, const AStarNode* _targetNode);
		
		/**
		* Save the given path of node ids as path of positions in the global path variable (m_foundPath)
		* @param vector<int>* the path of node ids to follow
		*/
		void savePositionPath(vector<int>* _nodeIdPath);
		
		/**
		* Returns the H cost of a given node.
		* @param int the node ID of the node, whose H should be computed
		* @return the H cost of the given node.
		*/
		int getHeuristic(const int _nodeId);
		
		/**
		* Returns the node from a given list with the lowest F cost.
		* @param AStarNodeMap* the map to get the node from
		* @return the node with the lowest F cost.
		*/
		AStarNode* getBestNodeFromOpenList(AStarNodeMap* _nodeList);
		
		/**
		* Returns whether the node with given node ID is in the map
		* @param AStarNodeMap* the map to search
		* @param int the node ID to search for
		* @return NULL if no node with that ID has been found, or the node if it has been found.
		*/
		AStarNode* isInList(AStarNodeMap* _nodeList, const int _nodeId);
		
		/**
		* Prints the given map
		* @param AStarNodeMap* the list to print
		*/
		void printList(AStarNodeMap* _nodeList);
		
		/**
		* Calculate the moving costs from the given node to the parent node
		* @param AStarNode* the node
		* @param AStarNode* the parent node
		* @return the costs.
		*/
		int getCosts(AStarNode* _node, AStarNode* _parent);
		
		/**
		 * Drives from parking lot to start position.
		 */
		void driveReverse();

		
	protected:
		/**
		* Rotate the given position around the given origin by the given angle
		* @param Position the position to rotate
		* @param Position the origin position around which we rotate
		* @param double the angle to rotate in radians
		* @return the rotated position.
		*/
		Position rotatePoint(Position _posToRotate, Position _rotateOrigin, double _angle);
		
		void clearMap(AStarNodeMap* _map);
};

#endif // ASTAR2DPLANER_H
