#include <iostream>
#include "Area.h"
#include "Environment.h"
#include "CVCoordDrawingArea.h"

extern Environment* g_Environment;
extern CVCoordDrawingArea* g_CoordDrawingArea;

//
// ----------------------------------------------------------------------------------------------------------
//

Area::Area(Car* _car, double _deltaT, double _width, double _height, double _angle, int _stepsWidth, int _stepsHeight, int _stepsAngle) {

	m_rows = _stepsHeight;
	m_columns = _stepsWidth;
	m_levels = _stepsAngle;
	
	m_width = _width;
	m_height = _height;
	m_angle = _angle;
	
	m_car = _car;
	m_deltaT = _deltaT;
	
	m_offset_rows = 0;
	m_offset_columns = 0;
	m_offset_levels = 0;
	
	m_widthStep = m_width / m_columns;
	m_heightStep = m_height / m_rows;
	m_angleStep = m_angle / m_levels;

	m_areaNodeMap = new map<int, AreaNode*>();
}

Area::~Area() {

	if(m_areaNodeMap != NULL) {
		AreaNodeMap::iterator i = m_areaNodeMap->begin();
		
		while(!m_areaNodeMap->empty()) {
			AreaNode* node = (*i).second;
			m_areaNodeMap->erase(i);
			
			if(node != NULL) {
				delete node;
			}
			
			i++;
		}
		
		delete m_areaNodeMap;
	}
}

//
// ----------------------------------------------------------------------------------------------------------
//

int Area::getNodeId(int _row, int _column, int _level) const {
	
	// Check if the node is still within the grid of this area
	if (!isValid(_row, _column, _level)) {
		return -1;
	}
	
	// row * m_cols + col + (m_rows * m_cols) * level
	return (_row * m_columns + _column + m_rows * m_columns * _level);
}

//
// ----------------------------------------------------------------------------------------------------------
//

int Area::getNumberOfNodes() const {

	return (m_rows * m_columns * m_levels);
}

//
// ----------------------------------------------------------------------------------------------------------
//

AreaNode* Area::getNode(int _nodeId) {

	// Search for the node
	AreaNodeMap::iterator i = m_areaNodeMap->find(_nodeId);
	
	// Check if we found the node
	if(i != m_areaNodeMap->end()) {
		return (*i).second;
	}
	
	return NULL;
}

//
// ----------------------------------------------------------------------------------------------------------
//

void Area::getSuccessors(int _nodeId, vector<AreaNode*>* _successors) {

	// Grab the area node associated to the given node id
	AreaNode* n = getNode(_nodeId);
	
	// Check if we have an area node
	if (n == NULL) {
		#ifdef AREA_DEBUG
			fprintf(stderr, "%s %d: The node %d is NULL.\n", __FILE__, __LINE__, _nodeId);
		#endif
		return;
	}
	
	// Check if the given node is an obstacle (an obstacle cannot have any successors)
	if (n->isObstacle()) {
		#ifdef AREA_DEBUG
			fprintf(stderr, "%s %d: The node %d is an obstacle.\n", __FILE__, __LINE__, _nodeId);
		#endif
		return;
	}
	
	// Get the reachable positions and test them for collision
	vector<Position> reachablePositions = m_car->getReachablePositions(m_deltaT, (n->getPosition()));
	vector<bool> collisionState = g_Environment->getCollisionState(reachablePositions);
	
	vector<Position>::iterator iterPos;
	vector<bool>::iterator iterCollision = collisionState.begin();

	for(iterPos = reachablePositions.begin(); iterPos != reachablePositions.end(); iterPos++) {

		// Create new AreaNodes for the received Positions
		AreaPosition areaPos = mapPositionToArea(&(*iterPos));
		
		// Check if the current reachable position is still within our area
		if(isValid(areaPos.row, areaPos.col, areaPos.level)) {
		
			//g_CoordDrawingArea->drawPoint((*iterPos).getX(), (*iterPos).getY(), 3, g_CoordDrawingArea->createColor("blue"));
			//g_CoordDrawingArea->updateWindow();
		
			int areaNodeId = getNodeId(areaPos.row, areaPos.col, areaPos.level);
			
			AreaNode* existingNode = getNode(areaNodeId);
			
			// Check if the area node already exists
			if(existingNode == NULL) {
			
				AreaNode* newNode = new AreaNode(areaNodeId, *(areaPos.pos), *iterCollision);
				_successors->push_back(newNode);

				// Make sure no node with that id exists, after that insert it
				m_areaNodeMap->erase(newNode->getNodeId());
				m_areaNodeMap->insert(AreaNodeMapPair(newNode->getNodeId(), newNode));
			} else {
			
				_successors->push_back(existingNode);
			}
		}
		
		iterCollision++;
	}
}

//
// ----------------------------------------------------------------------------------------------------------
//

bool Area::isValid(int _row, int _column, int _level) const {
	
	if (_row < 0 || _row >= m_rows) {
		#ifdef AREA_DEBUG
			fprintf(stderr, "%s %d: Row # %d is invalid.\n", __FILE__, __LINE__, _row);
		#endif
		return false;
	}
	if (_column < 0 || _column >= m_columns) {
		#ifdef AREA_DEBUG
			fprintf(stderr, "%s %d: Column # %d is invalid.\n", __FILE__, __LINE__, _column);
		#endif
		return false;
	}
	if (_level < 0 || _level >= m_levels) {
		#ifdef AREA_DEBUG
			fprintf(stderr, "%s %d: Level # %d is invalid.\n", __FILE__, __LINE__, _level);
		#endif
		return false;
	}
	
	return true;	
}

//
// ----------------------------------------------------------------------------------------------------------
//

bool Area::isValid(int _nodeId) const {
	
	if (_nodeId < 0) {
		return false;
	}
	if (_nodeId > getNodeId(m_rows-1, m_columns-1, m_levels-1)) {
		return false;
	}
	return true;	
}

//
// ----------------------------------------------------------------------------------------------------------
//

AreaPosition Area::mapPositionToArea(Position* _pos) {

	double x = _pos->getX();
	double y = _pos->getY();
	double theta = _pos->getTheta();
	
	// X-mapping
	double widthNeededSteps = x / m_widthStep;
	int gridCol = (int)(widthNeededSteps + 0.5);
	
	// Y-mapping
	double heightNeededSteps = y / m_heightStep;
	int gridRow = (int)(heightNeededSteps + 0.5);
	
	// Theta-mapping
	double angleNeededSteps = theta / m_angleStep;
	int gridLevel = (int)(angleNeededSteps + 0.5);
	
	// Fill the return struct
	AreaPosition aPos;
	aPos.pos = _pos;
	//aPos.mapPos = new Position(gridCol * m_widthStep, gridRow * m_heightStep, gridLevel * m_angleStep);
	aPos.col = gridCol + m_offset_columns;
	aPos.row = gridRow + m_offset_rows;
	aPos.level = gridLevel + m_offset_levels;
	
	if(aPos.level == m_levels) {
		aPos.level = 0;
	}
	
	return aPos;
}

//
// ----------------------------------------------------------------------------------------------------------
//

void Area::setGridOffset(int _rowOffset, int _colOffset, int _levelOffset) {

	m_offset_rows = ((m_rows/2) - _rowOffset);
	m_offset_columns = ((m_columns/2) - _colOffset);
	//m_offset_levels = -1;
	//m_offset_levels = ((m_levels/2) - _levelOffset);
}

int Area::getRowOffset() {
	return m_offset_rows;
}

int Area::getColumnOffset() {
	return m_offset_columns;
}

int Area::getLevelOffset() {
	return m_offset_levels;
}

//
// ----------------------------------------------------------------------------------------------------------
//

void Area::clearGridOffset() {
	m_offset_rows = 0;
	m_offset_columns = 0;
	m_offset_levels = 0;
}
