/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Master of Science - Computergraphics and Image Processing
 *
 * Robotics - Prof. Dr. Max Fischer
 * "Autonomous self-parking robot"
 * 
 * System: Linux, C++
 */

/**
 * USABILITY:
 *  Copyright (c) 2006 
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef AREA_H
#define AREA_H

/**
 * 'AreaNode' defines the nodes of the class 'Area', which
 * provides a map for the A* (A-Star) algorithm. The fields
 * of the map may be defined a accessible or as obstacles.
 *
 * @author Beatrix Müller, Thomas Wöllert
 */

#include <iostream>
#include <vector>
#include <string>
#include <map>

#include "AStarNode.h"
#include "Position.h"
#include "Car.h"

//#define AREA_DEBUG

using namespace std;

class AreaNode;

// Declaration of the map object to store the nodes in
typedef map<int, AreaNode*> AreaNodeMap;
typedef AreaNodeMap::value_type AreaNodeMapPair;

class AreaNode {

	private: 
		// The node id
		int  m_nodeId;	
		
		// The referenced position (x, y, angle)
		Position  m_pos;
		
		// If this node is an obstacle or not
		bool m_isObstacle;

	public:	
		/**
		* Default constructor
		*/
		AreaNode() {}
		
		/**
		* Constructor
		* @param int the node id
		* @param pos the position this node references in the real world
		* @param bool true if this node is an obstacle, false otherwise
		*/
		AreaNode(int _nodeId, Position _pos, bool _isObstacle) {
			m_nodeId     = _nodeId;
			m_pos 	     = _pos;
			m_isObstacle = _isObstacle;
		}
		
		/**
		* Default destructor
		*/
		virtual ~AreaNode() {}
		
		/**
		 * Returns the node ID.
		 * @return the node ID.
	         */
		int getNodeId() const {
			return m_nodeId;
		}
		
		/**
		* Returns the position this node references in the real world
		* @return the position
		*/
		Position getPosition() {
			return m_pos;
		}
		
		/**
		 * Returns whether the node is an obstacle
		 * @return true if the node is an obstacle, false otherwise
		 */
		bool isObstacle() const {
			return m_isObstacle;
		}
		
		/**
		 * Sets a node as obstacle or not as obstacle.
		 * @param bool true if the node should be set as an obstacle, false otherwise
		 */
		void setObstacle(bool _isObstacle) {
			m_isObstacle = _isObstacle;
		}
};

/**
* Struct holding an position and an associated row/col/level-pos in the
* A*-algorithm grid
*/
struct AreaPosition {

	// The position in the real world
	Position* pos;
	
	// The changed mapped position in the grid world
	Position* mapPos;
	
	// The row/col/level-position in the A*-grid
	int row;
	int col;
	int level;
};

class Area {

	private: 
		// The number of rows/columns/levels in the A*-grid
		int m_rows;
		int m_columns;
		int m_levels;
		
		// The offset of rows/columns/levels so our start node is in the mid of the grid
		int m_offset_rows;
		int m_offset_columns;
		int m_offset_levels;
		
		// The max width/height/angle of our real world area
		double m_width;
		double m_height;
		double m_angle;
		
		// Pre-calculated variables
		double m_widthStep;
		double m_heightStep;
		double m_angleStep;
		
		// The start and target node
		AreaNode* m_startNode;
		AreaNode* m_targetNode;
		
		// The node map
		AreaNodeMap* m_areaNodeMap;
		
		// The car model we are using to drive
		Car* m_car;
		
		// The deltaT time steps we are using
		double m_deltaT;

	public: 
		/**
		* Constructor
		* @param Car* the car model
		* @param double the deltaT time steps
		* @param double the width of the real area
		* @param double the height of the real area
		* @param double the maximum angle of the real area (normally dtor(360.0))
		* @param int the number of steps to use for the grid in the range [0 -> width]
		* @param int the number of steps to use for the grid in the range [0 -> height]
		* @param int the number of steps to use for the grid in the range [0 -> angle]
		*/
		Area(Car* _car, double _deltaT, double _width, double _height, double _angle, int _stepsWidth, int _stepsHeight, int _stepsAngle);
		
		/**
		* Destructor
		*/
		virtual ~Area();
		
		/**
		 * Returns the number of rows of the Area
		 * @return the number of rows
		 */
		int getRows() const {
			return m_rows;
		}
		
		/**
		 * Returns the number of columns of the Area
		 * @return the number of columns
		 */
		int getColumns() const {
			return m_columns;
		}
		
		/**
		 * Returns the number of levels of the Area
		 * @return the number of levels
		 */
		int getLevels() const {
			return m_levels;
		}
		
		/**
		 * Computes and returns the node ID based on the given A*-grid position
		 * @param int the row
		 * @param int the column
		 * @param int the level
		 * @return the node ID.
		 */
		int getNodeId(int _row, int _column, int _level) const;
		
		/**
		 * Returns the number of nodes int the Area
		 * @return the number of nodes
		 */
		int getNumberOfNodes() const;
		
		/**
		 * Returns the AreaNode
		 * @param int the node ID
		 * @return the AreaNode.
		 */	
		AreaNode* getNode(int _nodeId);
		
		/**
		 * Returns the row (in A*-grid-space) of a given node
		 * @param int the node ID
		 * @return the row of the given node
		 */
		int getRow(int _nodeId) {
			return (int)((_nodeId % (m_rows * m_columns)) / m_columns);
		}
		
		/**
		 * Returns the column (in A*-grid-space) of a given node
		 * @param int the node ID
		 * @return the colm_areaNodeVectorumn of the given node
		 */
		int getColumn(int _nodeId) {
			return (int)((_nodeId % (m_rows * m_columns)) % m_columns);
		}
		
		/**
		 * Returns the level (in A*-grid-space) of a given node
		 * @param int the node ID
		 * @return the level of the given node
		 */
		int getLevel(int _nodeId) {
			return (int)(_nodeId / (m_rows * m_columns));
		}
		
		/**
		 * Returns the ID of the start node
		 * @return the start node ID
		 */
		int getStartNodeId() const {
			return m_startNode->getNodeId();
		}
		
		/**
		 * Returns the ID of the target node
		 * @return the target node ID
		 */
		int getTargetNodeId() const{
			return m_targetNode->getNodeId();
		}
		
		/**
		 * Returns the successors of a node, which are its adjacent nodes
		 * @param int the node ID to get the adjacent nodes from
		 * @param vector<AreaNode>& the vector with the successors to return
		 */
		void Area::getSuccessors(int _nodeId, vector<AreaNode*>* _successors);
		
		/**
		 * Sets the start node
		 * @param AreaNode* the start node
		 */
		void setStartNode(AreaNode* _startNode) {
			
			// Remember the start node and insert it into the area node map
			m_startNode = _startNode;
			m_areaNodeMap->insert(AreaNodeMapPair(m_startNode->getNodeId(), m_startNode));
		}
		
		/**
		 * Sets the target node
		 * @param AreaNode* the target node
		 */
		void setTargetNode(AreaNode* _targetNode) {
		
			// Remember the target node and insert it into the area node map
			m_targetNode = _targetNode;
			m_areaNodeMap->insert(AreaNodeMapPair(m_targetNode->getNodeId(), m_targetNode));
		}
		
		/**
		* Map a given position in the real world into the area-grid
		* @param Position* the position
		* @return the mapped area position
		*/
		AreaPosition mapPositionToArea(Position* _pos);
		
		/**
		* Set the grid offset to use in order to set the start node in the mid of the grid
		* @param int the row offset
		* @param int the column offset
		* @param int the level offset
		*/
		void setGridOffset(int _rowOffset, int _colOffset, int _levelOffset);
	
		int getRowOffset();
		int getColumnOffset();	
		int getLevelOffset();
	
		/**
		* Reset the grid offset to (0/0/0)
		*/
		void clearGridOffset();

	private:
		/**
		 * Returns whether the row and column are within the range of the Area
		 * @param int the row
		 * @param int the column
		 * @param int the level
		 * @return true if the node is valid, false otherwise.
		 */
		bool isValid(int _row, int _column, int _level) const;
		
	       /**
		* Returns whether the node ID is within the range of the Area
		* @param int the node ID
		* @return true if the node is valid, false otherwise.
		*/
		bool isValid(int nodeId) const;
};

#endif
