#include "CVCoordDrawingArea.h"
#include "Global.h"
#include "Position.h"

#include <iostream>
#include "math.h"

/*
*-------------------------------------------------------------------------------------------------------
*/
CVCoordDrawingArea::CVCoordDrawingArea(double _minX, double _maxX, double _minY, double _maxY, int _width, int _height) {
	minX = _minX;
	maxX = _maxX;
	minY = _minY;
	maxY = _maxY;
	
	// Set the image size
	imgSize.width = _width; 
  	imgSize.height = _height;
  	
  	// Create the back and front image
  	backImage = cvCreateImage(imgSize, IPL_DEPTH_8U, 3);
  	
  	// Semaphore init
  	sem_init(&backImageSemaphore, 0, 1);
  	sem_init(&frontImageSemaphore, 0, 1);
  	
  	// Add colors
  	Color colorRed;
  	colorRed.name = "red";
  	colorRed.value = CV_RGB(255,0,0);
  	colors.push_back(colorRed);
  	
  	Color colorGreen;
  	colorGreen.name = "green";
  	colorGreen.value = CV_RGB(0,255,0);
  	colors.push_back(colorGreen);
  	
  	Color colorBlue;
  	colorBlue.name = "blue";
  	colorBlue.value = CV_RGB(0,0,255);
  	colors.push_back(colorBlue);
  	
  	Color colorBlack;
  	colorBlack.name = "black";
  	colorBlack.value = CV_RGB(0,0,0);
  	colors.push_back(colorBlack);
  	
  	Color colorWhite;
  	colorWhite.name = "white";
  	colorWhite.value = CV_RGB(255,255,255);
  	colors.push_back(colorWhite);
  	
  	Color colorGray;
  	colorGray.name = "lightgray";
  	colorGray.value = CV_RGB(230,230,230);
  	colors.push_back(colorGray);
  	
  	// Init the font to use
	cvInitFont(&font, CV_FONT_VECTOR0, 0.4, 0.4, 0.0, 1);
	
	// Init the window information
	windowHandle = -1;
}

CVCoordDrawingArea::~CVCoordDrawingArea() {

	// Free the image memory
	cvReleaseImage(&backImage);
	
	if(frontImage != 0) {
		cvReleaseImage(&frontImage);
	}
}

/*
*-------------------------------------------------------------------------------------------------------
*/
CvSize CVCoordDrawingArea::getImageSize() {
	return imgSize;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
int CVCoordDrawingArea::showWindow(const char* _name, unsigned long _flags) {
	if(windowHandle == -1) {
		windowHandle = cvNamedWindow(_name, _flags);
		windowName = new char[255];
		strcpy(windowName, _name);
		
		updateWindow();
	}
	
	return windowHandle;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::updateWindow() {
	if(windowHandle != -1) {
	
		sem_wait(&frontImageSemaphore);
				
		cvShowImage(windowName, backImage);
		
		sem_post(&frontImageSemaphore);
	}
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::clear() {
	setBackgroundColor(createColor("white"));
}

/*
*-------------------------------------------------------------------------------------------------------
*/
CvScalar CVCoordDrawingArea::createColor(unsigned char _red, unsigned char _green, unsigned char _blue) {
	return CV_RGB(_red, _green, _blue);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
CvScalar CVCoordDrawingArea::createColor(string _name) {

	for(unsigned int i = 0; i < colors.size(); i++) {
		if(colors[i].name == _name)
			return colors[i].value;
	}
	
	CvScalar empty;
	return empty;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::setBackgroundColor(CvScalar _color) {

	sem_wait(&backImageSemaphore);
	cvRectangle(backImage, cvPoint(0,0), cvPoint(imgSize.width,imgSize.height), _color, CV_FILLED);
	sem_post(&backImageSemaphore);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::drawLine(double _x1, double _y1, double _x2, double _y2, int _lineWidth, CvScalar _color) {

	CvPoint pos1 = coordinatesToPixels(_x1, _y1);
	CvPoint pos2 = coordinatesToPixels(_x2, _y2);
	
	sem_wait(&backImageSemaphore);
	cvLine(backImage, pos1, pos2, _color, _lineWidth);
	sem_post(&backImageSemaphore);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::drawRect(double _x1, double _y1, double _x2, double _y2, int _lineWidth, CvScalar _color) {

	CvPoint pos1 = coordinatesToPixels(_x1, _y1);
	CvPoint pos2 = coordinatesToPixels(_x2, _y2);
	
	sem_wait(&backImageSemaphore);
	cvRectangle(backImage, pos1, pos2, _color, _lineWidth);
	sem_post(&backImageSemaphore);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::drawPoint(double _x, double _y, int _thickness, CvScalar _color) {

	// Validate the coordinates
	if(!validate(_x, _y)) {
		return;
	}
	
	CvPoint pos = coordinatesToPixels(_x, _y);

	sem_wait(&backImageSemaphore);
	cvLine(backImage, pos, pos, _color, _thickness);
	sem_post(&backImageSemaphore);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::drawText(double _x, double _y, string _text, CvScalar _color) {

	CvPoint pos = coordinatesToPixels(_x, _y);

	sem_wait(&backImageSemaphore);
	cvPutText(backImage, _text.c_str(), pos, &font, _color);
	sem_post(&backImageSemaphore);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::drawRotRect(double _xCenter, double _yCenter, double _angle, double _width, double _height, int _lineWidth, CvScalar _color) {
	
	//
	// Calculate the rectangle points
	//
	
	CvDoublePoint centerPoint;
	centerPoint.x = _xCenter;
	centerPoint.y = _yCenter;
	
	// Top-left corner
	CvDoublePoint pTopLeft;
	pTopLeft.x = _xCenter - (_width/2.0);
	pTopLeft.y = _yCenter + (_height/2.0);
	
	// Top-right corner
	CvDoublePoint pTopRight;
	pTopRight.x = _xCenter + (_width/2.0);
	pTopRight.y = _yCenter + (_height/2.0);
	
	// Bottom-left corner
	CvDoublePoint pBotLeft;
	pBotLeft.x = _xCenter - (_width/2.0);
	pBotLeft.y = _yCenter - (_height/2.0);
	
	// Bottom-right corner
	CvDoublePoint pBotRight;
	pBotRight.x = _xCenter + (_width/2.0);
	pBotRight.y = _yCenter - (_height/2.0);
	
	// Transform the points based on the given angle
	pTopLeft = rotatePoint(pTopLeft, centerPoint, _angle);
	pTopRight = rotatePoint(pTopRight, centerPoint, _angle);
	pBotLeft = rotatePoint(pBotLeft, centerPoint, _angle);
	pBotRight = rotatePoint(pBotRight, centerPoint, _angle);
	
	// Draw the rectangle
	drawLine(pTopLeft.x, pTopLeft.y, pTopRight.x, pTopRight.y, _lineWidth, _color);
	drawLine(pTopRight.x, pTopRight.y, pBotRight.x, pBotRight.y, _lineWidth, _color);
	drawLine(pBotRight.x, pBotRight.y, pBotLeft.x, pBotLeft.y, _lineWidth, _color);
	drawLine(pBotLeft.x, pBotLeft.y, pTopLeft.x, pTopLeft.y, _lineWidth, _color);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::drawRobot(Position _robotPos, RobotDimension _dim, int _lineWidth, CvScalar _color) {

	// Draw the robot center
	drawPoint(_robotPos.getX(), _robotPos.getY(), _lineWidth, _color);
	
	// Calc the real rectangle center
	double rectCenterX = _robotPos.getX() + _dim.xOffset;
	double rectCenterY = _robotPos.getY() + _dim.yOffset;
	
	// Draw the rotated rectangle of the robot
	drawRotRect(rectCenterX, rectCenterY, _robotPos.getTheta(), _dim.length, _dim.width, _lineWidth, _color);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::drawParkingLines(Point* _lines, int _count, int _lineWidth, CvScalar _color) {

	if(_lines == NULL) {
		return;
	}

	Point startPoint = _lines[0];
	Point endPoint;

	for(int i = 1; i < _count; i++) {
	
		endPoint = _lines[i];
		drawLine(startPoint.x, startPoint.y, endPoint.x, endPoint.y, _lineWidth, _color);
		startPoint = endPoint;
	}
}

/*
*-------------------------------------------------------------------------------------------------------
*/
bool CVCoordDrawingArea::validate(double _x, double _y) {

	// Check x-coordinate
	if(_x < minX || _x > maxX) {
		return false;
	}
	
	// Check y-coordinate
	if(_y < minY || _y > maxY) {
		return false;
	}
	
	return true;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
CvPoint CVCoordDrawingArea::coordinatesToPixels(double _xCoord, double _yCoord) {

	CvPoint pos;

	// Transform the x-coordinate into pixels
	double xLength = maxX + abs(minX);
	double xPixelsPerUnit = (double)(imgSize.width) / xLength;
	double xCoordDiff = abs(minX - _xCoord);	
	int xPixel = (int)(xPixelsPerUnit * xCoordDiff);
	
	// Transform the y-coordinate into pixels
	double yLength = maxY + abs(minY);
	double yPixelsPerUnit = (double)(imgSize.height) / yLength;
	double yCoordDiff = abs(minY - _yCoord);
	int yPixel = (imgSize.height - (int)(yPixelsPerUnit * yCoordDiff));	
	
	pos.x = xPixel;
	pos.y = yPixel;
	
	return pos;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
CvDoublePoint CVCoordDrawingArea::rotatePoint(CvDoublePoint _point, CvDoublePoint _rotationOrigin, double _angle) {

	_rotationOrigin.x = -1.0*_rotationOrigin.x;
	_rotationOrigin.y = -1.0*_rotationOrigin.y;
	
	CvDoublePoint transPoint;
	transPoint.x = _point.x + _rotationOrigin.x;
	transPoint.y = _point.y + _rotationOrigin.y;
	
	double newX = (transPoint.x * cos(_angle)) - (transPoint.y * sin(_angle));
	double newY = (transPoint.x * sin(_angle)) + (transPoint.y * cos(_angle));
	
	transPoint.x = newX;
	transPoint.y = newY;
	
	_rotationOrigin.x = -1.0*_rotationOrigin.x;
	_rotationOrigin.y = -1.0*_rotationOrigin.y;
	
	CvDoublePoint rotatedPoint;
	rotatedPoint.x = transPoint.x + _rotationOrigin.x;
	rotatedPoint.y = transPoint.y + _rotationOrigin.y;
	
	return rotatedPoint;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::drawGrid(double _xCoverage, double _yCoverage, int _lineWidth, CvScalar _color) {

	// Draw the x-lines
	double currentX = minX;
	
	while(currentX <= maxX) {
	
		drawLine(currentX, minY, currentX, maxY, _lineWidth, _color);
		currentX = currentX + _xCoverage;
	}
	
	// Draw the y-lines
	double currentY = minY;
	
	while(currentY <= maxY) {
	
		drawLine(minX, currentY, maxX, currentY, _lineWidth, _color);
		currentY = currentY + _yCoverage;
	}
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void CVCoordDrawingArea::drawGridText(double _xCoverage, double _yCoverage, CvScalar _color) {

	// Draw the x-line texts
	double currentX = minX;
	
	while(currentX <= maxX) {
	
		drawText(currentX, -0.15, TtoString(currentX), _color);
		currentX = currentX + _xCoverage;
	}
	
	// Draw the y-line texts
	double currentY = minY;
	
	while(currentY <= maxY) {
	
		if(currentY != 0.0) {
			drawText(-0.2, (currentY+0.1), TtoString(currentY), _color);
		}
		currentY = currentY + _yCoverage;
	}
}
