/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Master of Science - Computergraphics and Image Processing
 *
 * Robotics - Prof. Dr. Max Fischer
 * "Autonomous self-parking robot"
 * 
 * System: Linux, C++
 */

/**
 * USABILITY:
 *  Copyright (c) 2006 
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * Class Description:
 *	Wrapper for the DrawingArea class to present the image as a coordinate
 *	system where you can draw into the system with real coordinates instead of pixels
 *
 * @author Thomas Wöllert
 */

#ifndef CV_COORDDRAWINGAREA_H
#define CV_COORDDRAWINGAREA_H

#include "Robot.h"
#include "collision_test.h"

#include "cv.h"
#include "highgui.h"
#include <string>
#include <vector>
#include <semaphore.h>

using namespace std;

class Position;

struct Color {
	string name;
	CvScalar value;
};

struct CvDoublePoint {
	double x;
	double y;
};

class CVCoordDrawingArea {

	protected:
		// The backbuffer image we are drawing into
		IplImage* backImage;
		
		// The image shown on the screen (copy of the backbuffer)
		IplImage* frontImage;
		
		// The size of the image
		CvSize imgSize;
		
		// Color management
		vector<Color> colors;
		
		// The font to use
		CvFont font;
	
		// The x-coordinate range
		double minX;
		double maxX;
		
		// The y-coordinate range
		double minY;
		double maxY;
		
		// Window information
		int windowHandle;
		char* windowName;
		
		// Locks
		sem_t backImageSemaphore;
		sem_t frontImageSemaphore;
	
	public:
		/**
		* Constructor
		* @param double min x-value
		* @param double max x-value
		* @param double min y-value
		* @param double max y-value
		* @param int the width of the image in pixels
		* @param int the height of the image in pixels
		*/
		CVCoordDrawingArea(double _minX, double _maxX, double _minY, double _maxY, int _width, int _height);
	
		/**
		* Destructor
		*/
		virtual ~CVCoordDrawingArea();
		
		/**
		* Get the size of the image
		* @return the size
		*/
		CvSize getImageSize();
		
		/**
		* Show the window
		* @param const_char* the title of the window
		* @param ulong the flags
		* @return the window handle
		*/
		int showWindow(const char* _name, unsigned long _flags);
		
		/**
		* Update the window
		*/
		void updateWindow();
		
		/**
		* Clear the window
		*/
		void clear();
		
		/**
		* Create a color
		* @param uchar red (0-255)
		* @param uchar green (0-255)
		* @param uchar blue (0-255)
		* @return the color
		*/
		CvScalar createColor(unsigned char _red, unsigned char _green, unsigned char _blue);
		
		/**
		* Create a pre-defined color
		* @param string the name of the color
		* @return the color
		*/
		CvScalar createColor(string _name);
		
		/**
		* Set the background color of the drawing area
		* @param CvScalar the color to use (use createColor(...) or CV_RGB(uchar, uchar, uchar))
		*/
		void setBackgroundColor(CvScalar _color);
		
		/**
		* Draw a line
		* @param double the start-x
		* @param double the start-y
		* @param double the end-x
		* @param double the end-y
		* @param int the linewidth (in px)
		* @param CvScalar the color (use createColor(...) or CV_RGB(uchar, uchar, uchar))
		*/	
		void drawLine(double _x1, double _y1, double _x2, double _y2, int _lineWidth, CvScalar _color);
		
		/**
		* Draw a rectangle
		* @param double the left-upper-x
		* @param double the left-upper-y
		* @param double the right-lower-x
		* @param double the right-lower-y
		* @param int the linewidth (in px)
		* @param CvScalar the color (use createColor(...) or CV_RGB(uchar, uchar, uchar))
		*/
		void drawRect(double _x1, double _y1, double _x2, double _y2, int _lineWidth, CvScalar _color);
		
		/**
		* Draw a text
		* @param double the left-upper-x
		* @param double the left-upper-y
		* @param string the text to draw
		* @param CvScalar the color (use createColor(...) or CV_RGB(uchar, uchar, uchar))
		*/
		void drawText(double _x, double _y, string _text, CvScalar _color);
		
		/**
		* Draw a point
		* @param double the point-x
		* @param double the point-y
		* @param int the thickness (in px)
		* @param CvScalar the color (use createColor(...) or CV_RGB(uchar, uchar, uchar))
		*/
		void drawPoint(double _x, double _y, int _thickness, CvScalar _color);
		
		/**
		* Draw the outline of the robot at the given position
		* @param Position the robot position
		* @param RobotDimension the robot size
		* @param int the thickness (in px)
		* @param CvScalar the color (use createColor(...) or CV_RGB(uchar, uchar, uchar))
		*/
		void drawRobot(Position _robotPos, RobotDimension _dim, int _lineWidth, CvScalar _color);
		
		void drawParkingLines(Point* _lines, int _count, int _lineWidth, CvScalar _color);
		
		/**
		* Draw a rotated rectangle
		* @param double rotation center-x
		* @param double rotation center-y
		* @param double the angle to rotate
		* @param double the width of the rectangle
		* @param double the height of the rectangle
		* @param int the thickness (in px)
		* @param CvScalar the color (use createColor(...) or CV_RGB(uchar, uchar, uchar))
		*/
		void drawRotRect(double _xCenter, double _yCenter, double _angle, double _width, double _height, int _lineWidth, CvScalar _color);
		
		/**
		* Draw the grid
		* @param double the x-range covered by one grid element
		* @param double the y-range covered by one grid element
		* @param int the line width
		* @param CvScalar the color to draw with (use createColor(...) or CV_RGB(uchar, uchar, uchar))
		*/
		void drawGrid(double _xCoverage, double _yCoverage, int _lineWidth, CvScalar _color);
		
		/**
		* Draw the grid text
		* @param double the x-range covered by one grid element
		* @param double the y-range covered by one grid element
		* @param CvScalar the color to use (use createColor(...) or CV_RGB(uchar, uchar, uchar))
		*/
		void drawGridText(double _xCoverage, double _yCoverage, CvScalar _color);
		
	protected:
		/**
		* Check if the specified point is still within the
		* area of the coordinate system
		* @param double the x-coordinate
		* @param double the y-coordinate
		* @return true if the pixel is within the area, false otherwise
		*/
		bool validate(double _x, double _y);
	
		/**
		* Transform the given real coordinates into pixel values
		* @param double the x-coordinate
		* @param double the y-coordinate
		* @return the pixel position
		*/
		CvPoint coordinatesToPixels(double _xCoord, double _yCoord);
		
		/**
		* Rotate a given point around a given rotation origin by the given angle
		* @param CvDoublePoint the point to rotate
		* @param CvDoublePoint the rotation origin
		* @param double the angle
		* @return the new rotated point
		*/
		CvDoublePoint rotatePoint(CvDoublePoint _point, CvDoublePoint _rotationOrigin, double _angle);
};

#endif // CV_COORDDRAWINGAREA_H
