////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// File:	car.cpp
// Author:	Stefan Hertel
// Date:	03.04.2006
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "Car.h"
#include "Robot.h"
#include "Config.h"
#include "math.h"

#define CAR_EPSILON 1e-8

//! creates a new car
Car::Car(Robot* _robot, double _maxSteeringAngle, double _px, double _nx, double _py, double _ny, double _wheelbase) {

	m_robot = _robot;
	m_lengthPX = _px;
	m_lengthNX = _nx;
	m_widthPY = _py;
	m_widthNY = _ny;
	m_wheelbase = _wheelbase;
	m_maxSteeringAngle = _maxSteeringAngle;
}

// destructor
//! destroyes this car
Car::~Car()
{
}

//! Retrieve all reachable positions (from the current position)
// \param double, deltaT we want to drive to these positions
// \param Position, the current Position
// \return std:vector<Position>, list of all reachable positions
std::vector<Position> Car::getReachablePositions( double deltaT, Position current )
{
	double driveSpeed = m_robot->getDriveSpeed();

	std::vector<Position> retVec(6);
	// drive forwards to the left 
	retVec[0] = getNextPosition( deltaT, current, 0., 0., driveSpeed, m_maxSteeringAngle );
	// drive forwards
	retVec[1] = getNextPosition( deltaT, current, 0., 0., driveSpeed, 0. );
	// drive forwards to the right
	retVec[2] =  getNextPosition( deltaT, current, 0., 0., driveSpeed, -1.0*m_maxSteeringAngle );
	// drive backwards to the left
	retVec[3] =  getNextPosition( deltaT, current, 0., 0., -1.0*driveSpeed, m_maxSteeringAngle );
	// drive backwards
	retVec[4] =  getNextPosition( deltaT, current, 0., 0., -1.0*driveSpeed, 0. );
	// drive backwards to the right
	retVec[5] =  getNextPosition( deltaT, current, 0., 0., -1.0*driveSpeed, -1.0*m_maxSteeringAngle );
	// return the vector
	return retVec;
}

//! Set a list of route commands the car needs to follow.
//! Start following the commands immediatly. Method returns when
//! the car stopped at the final destination.
//! \param const std:list<RouteCommand>&, the list of route commands
void Car::setListOfRouteCommands( const std::list<RouteCommand>& cmdList )
{
	m_robot->setListOfRouteCommands( cmdList );
}

void Car::drive(ENUMdriveDirection _direction, ENUMdriveSteering _steering) {

	double driveSpeed = 0.0;
	double turnRate = 0.0;

	// Set the proper drive speed
	switch(_direction) {

		case driveDirectionForward:
			driveSpeed = m_robot->getDriveSpeed();
		break;

		case driveDirectionBackward:
			driveSpeed = -1.0*(m_robot->getDriveSpeed());
		break;

		default:
			driveSpeed = 0.0;
	}

	// Set the proper turn rate
	switch(_steering) {

		case driveSteeringStraight:
			turnRate = 0.0;
		break;

		case driveSteeringLeft:
			turnRate = m_maxSteeringAngle;
		break;

		case driveSteeringRight:
			turnRate = -1.0*m_maxSteeringAngle;
		break;

		default:
			turnRate = 0.0;
	}

	m_robot->move(driveSpeed, turnRate);
}

void Car::drivePath(double _deltaT, vector<Position*>* path) {

	list<RouteCommand> cmdList;
	Position currentPos = m_robot->getCurrentPosition();
	Position nextPos;

	for(unsigned int i = 0; i < path->size(); i++) {
	
		nextPos = *(path->at(i));
		
		// Check if we got a real new point
		if(nextPos.getX() != currentPos.getX() && nextPos.getY() != currentPos.getY() && nextPos.getTheta() != currentPos.getTheta()) {
		
			// Determine the turn radius
			double deltaTheta = nextPos.getTheta() - currentPos.getTheta();
			
			// Theta range is from 0->2*PI, so correct the result
			if(deltaTheta > M_PI) {
				deltaTheta -= 2*M_PI;
			} else if(deltaTheta < -M_PI) {
				deltaTheta += 2*M_PI;
			}

			// Determine the drive direction		
			Position rotNextPos = rotatePoint(nextPos, currentPos, -1.0*currentPos.getTheta());	
			ENUMdriveDirection direction;
	
			if(rotNextPos.getX() >= currentPos.getX()) {
	
				// Drive forward
				direction = driveDirectionForward;
	
			} else {
	
				// Drive backward
				direction = driveDirectionBackward;
			}
	
			// Create the route command
			RouteCommand cmd = RouteCommand(direction, _deltaT, deltaTheta);
			m_robot->driveCarLikeTo(nextPos, cmd, DRIVE_TO_TARGET_X_THRESHOLD, DRIVE_TO_TARGET_Y_THRESHOLD, DRIVE_TO_TARGET_THETA_THRESHOLD);
		}

		currentPos = m_robot->getCurrentPosition();
	}
	
	m_robot->stop();
}

Position Car::rotatePoint(Position _posToRotate, Position _rotateOrigin, double _angle) {

	_rotateOrigin.setX(-1.0*_rotateOrigin.getX());
	_rotateOrigin.setY(-1.0*_rotateOrigin.getY());

	Position transPoint;
	transPoint.setX(_posToRotate.getX() + _rotateOrigin.getX());
	transPoint.setY(_posToRotate.getY() + _rotateOrigin.getY());
	
	double newX = (transPoint.getX() * cos(_angle)) - (transPoint.getY() * sin(_angle));
	double newY = (transPoint.getX() * sin(_angle)) + (transPoint.getY() * cos(_angle));
	
	transPoint.setX(newX);
	transPoint.setY(newY);
	
	_rotateOrigin.setX(-1.0*_rotateOrigin.getX());
	_rotateOrigin.setY(-1.0*_rotateOrigin.getY());

	Position rotatedPoint = Position();
	rotatedPoint.setX(transPoint.getX() + _rotateOrigin.getX());
	rotatedPoint.setY(transPoint.getY() + _rotateOrigin.getY());
	rotatedPoint.setTheta(0.0);
	
	return rotatedPoint;
}

void Car::stop() {
	m_robot->stop();
}

// get the position when driving a specified time with the same speed and steering angle
// \param deltaT the time to drive
// \param initialPosition the initial Postition
// \param initialSpeed the initial speed ( currently unused )
// \param initialSteeringAngle the initial steering angle ( currently unused )
// \param speed the speed to drive with
// \param steeringAngle the steering angle to drive with
Position Car::getNextPosition( double deltaT, Position initialPosition, double initialSpeed, double initialSteeringAngle, double speed, double steeringAngle ) const
{
	// twoellert
	double initialPositionTheta = initialPosition.getTheta();
	if(initialPositionTheta > M_PI) {
		initialPositionTheta -= 2*M_PI;
	}

	// avoid division trough zero
	if( fabs(steeringAngle) < CAR_EPSILON )
	{
		double newX = initialPosition.getX() + speed * cos( initialPositionTheta ) * deltaT;
		double newY = initialPosition.getY() + speed * sin( initialPositionTheta ) * deltaT;
		// a steering angle of 0 => no change in theta angle
		return Position( newX, newY, initialPosition.getTheta() );
	}
	//double theta = (speed/getWheelbase()) * tan( steeringAngle ) * deltaT;
	double theta = (speed/getWheelbase()) * tan( steeringAngle * deltaT);

	double tanPhi = tan( steeringAngle );
	double newTheta = initialPositionTheta + theta;
	double newX = initialPosition.getX() + (getWheelbase()/ tanPhi ) * (sin( initialPositionTheta + theta ) - sin( initialPositionTheta ) );
	double newY = initialPosition.getY() - (getWheelbase()/ tanPhi ) * (cos( initialPositionTheta + theta ) - cos( initialPositionTheta ) );
	
	// Make sure to avoid negative theta (real simulation theta: 0->2*PI) (twoellert)
 	if(newTheta < 0) {
 		newTheta += 2*M_PI;
 	}
	
	return Position( newX, newY, newTheta );
}

// get the minimum turning radius
double Car::getMinimumTurningRadius() const
{
	return getWheelbase()/m_maxSteeringAngle;
}
