////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// File:	car.h
// Author:	Stefan Hertel
// Date:	03.04.2006
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef CAR_H
#define CAR_H

#include <vector>
#include <list>

#include "Position.h"
#include "RouteCommand.h"
#include "Robot.h"

//! \brief	this class stores all parameters used by a parking simulation to descibe a car model
//!			a car has a own coordinate system with its origin in the center of the rear axis
//!			all distances and angles are given in this system
class Car
{
public:
	// constructor 
	//! creates a new car
	/**
	* Constructs a new car
	* @param Robot* the robot wrapped by this car
	* @param double the maximum steering angle for the simulated car (in rad)
	* @param double (px) the length of the car from the origin to the front (direction positive x) (in meter)
	* @param double (nx) the length of the car from the origin to the back (direction negative y) (in meter)
	* @param double (py) the width of the car from the origin to the left (direction positive x) (in meter)
	* @param double (ny) the width of the car from the origin to the right (direction negative y) (in meter)
	* @param double (wheelbase) the distance from the front axis (in meter)
	*/
	Car(Robot* _robot, double _maxSteeringAngle, double _px, double _nx, double _py, double _ny, double _wheelbase);

	// destructor
	//! destroyes this car
	virtual ~Car();

	// set the length of the car from origin to the front (direction positive x)
	// \param the new length in meter
	void setLengthPX( double lengthPX ) { m_lengthPX = lengthPX; }

	// set the length of the car from origin to the back (direction negative x)
	// \param the new length in meter
	void setLengthNX( double lengthNX ) { m_lengthNX = lengthNX; }

	// set the width of the car from origin to the left (direction positive y)
	// \param the new width in meter
	void setWidthPY( double widthPY ) { m_widthPY = widthPY; }

	// set the width of the car from origin to the right (direction negative y)
	// \param the new width in meter
	void setWidthNY( double widthNY ) { m_widthNY = widthNY; }

	// set the distance of the front axis
	// \param the new wheelbase in meter
	void setWheelbase( double wheelbase ) { m_wheelbase = wheelbase; }
	
	// get the length of the car from origin to the front (direction positive x)
	double getLengthPX() const { return m_lengthPX; }

	// get the length of the car from origin to the back (direction negative x)
	double getLengthNX() const { return m_lengthNX; }

	// get the width of the car from origin to the left (direction positive y)
	double getWidthPX() const { return m_widthPY; }

	// get the width of the car from origin to the right (direction negative y)
	double getWidthNY() const { return m_widthNY; }

	// get the distance of the front axis
	double getWheelbase() const { return m_wheelbase; }

	// get the minimum turning radius
	double getMinimumTurningRadius() const;

	//! Retrieve all reachable positions (from the current position)
	// \param double, deltaT we want to drive to these positions
	// \param Position, the current Position
	// \return std:vector<Position>, list of all reachable positions
	std::vector<Position> getReachablePositions( double deltaT, Position current );

	//! Set a list of route commands the car needs to follow.
	//! Start following the commands immediatly. Method returns when
	//! the car stopped at the final destination.
	//! \param const std:list<RouteCommand>&, the list of route commands
	void setListOfRouteCommands( const std::list<RouteCommand>& cmdList );

	/**
	* Manual drive (drives into the given direction till a new direction is set or stop is called)
	* @param ENUMdriveDirection the direction to drive into
	* @param ENUMdriveSteering the steering to use
	*/
	void drive(ENUMdriveDirection _direction, ENUMdriveSteering _steering);

	void drivePath(double _deltaT, vector<Position*>* path);

	Position rotatePoint(Position _posToRotate, Position _rotateOrigin, double _angle);

	/**
	* Stops any movement of the car
	*/
	void stop();

protected:

	// get the position when driving a specified time with the same speed and steering angle
	// \param deltaT the time to drive
	// \param initialPosition the initial Postition
	// \param initialSpeed the initial speed
	// \param initialSteeringAngle the initial steering angle 
	// \param speed the speed to drive with
	// \param steeringAngle the steering angle to drive with
	Position getNextPosition( double deltaT, Position initialPosition, double initialSpeed, double initalSteeringAngle, double speed, double steeringAngle ) const;

private:
	friend class UserPlaner;

	// the fixed parameters
	Robot* m_robot;

	// the length of the car from origin to the front (direction positive x)
	double m_lengthPX;

	// the lenght of the car from origin to the back (direction negative x)
	double m_lengthNX;

	// the width of the car from origin to the left (direction positive y)
	double m_widthPY;

	// the width of the car from origin to the right (direction negative y)
	double m_widthNY;

	// the distance of the front axis
	double m_wheelbase;

	// The given max steering angle of this robot
	double m_maxSteeringAngle;

	// the variable parameters

};

#endif // CAR_H
