#include "Environment.h"
#include "collision_test.h"
#include "cv.h"
#include "FindParkingLot.h"
#include "Config.h"

/*
*-------------------------------------------------------------------------------------------------------
*/
Environment::Environment(Robot* _robot, double _width, double _height, double _angle) {
	m_robot = _robot;
	m_areaWidth = _width;
	m_areaHeight = _height;
	m_areaAngle = _angle;
}

Environment::~Environment() {
	if(parkingLines != NULL) {
		delete[] parkingLines;
	}
}

/*
*-------------------------------------------------------------------------------------------------------
*/
double Environment::getAreaWidth() {
	return m_areaWidth;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
double Environment::getAreaHeight() {
	return m_areaHeight;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
double Environment::getAreaAngle() {
	return m_areaAngle;
}

Point* Environment::getParkingLines() {

	return parkingLines;
}

void Environment::setParkingLines(Point leftBorderPoint, Point leftTopPoint, Point leftBottomPoint, Point rightBottomPoint, Point rightTopPoint, Point rightBorderPoint) {

	parkingLines = new Point[6];
	parkingLines[5].x = -leftBorderPoint.x;
	parkingLines[5].y = -leftBorderPoint.y;
	
	parkingLines[4].x = -leftTopPoint.x;
	parkingLines[4].y = -leftTopPoint.y;
	
	parkingLines[3].x = -leftBottomPoint.x;
	parkingLines[3].y = -leftBottomPoint.y;
	
	parkingLines[2].x = -rightBottomPoint.x;
	parkingLines[2].y = -rightBottomPoint.y;
	
	parkingLines[1].x = -rightTopPoint.x;
	parkingLines[1].y = -rightTopPoint.y;
	
	parkingLines[0].x = -rightBorderPoint.x;
	parkingLines[0].y = -rightBorderPoint.y;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
vector<bool> Environment::getCollisionState(const vector<Position>& _posList) {

	// Go through all requested positions and check them for collision
	vector<bool> collisions = vector<bool>();
		
	RobotDimension dim = m_robot->getRobotDimension();
	
	// Add the ensurance thresholds
	dim.width += COLLISION_STATE_WIDTH_THRESHOLD;
	dim.length += COLLISION_STATE_LENGTH_THRESHOLD;

	for(vector<Position>::const_iterator iter = _posList.begin(); iter != _posList.end(); iter++) {
	
		// Check the current position
		
		// Calc the real rectangle center
		double rectCenterX = (*iter).getX() + dim.xOffset;
		double rectCenterY = (*iter).getY() + dim.yOffset;
		Point* robotRect = rotateRectangle(rectCenterX, rectCenterY, (*iter).getTheta(), dim.length, dim.width);
		
		int numberOfCollisions = rectline_collision(robotRect, parkingLines, 6);
		
		if(numberOfCollisions > 0) {
			// Collision
			collisions.push_back(true);
		} else {
			// No collision
			collisions.push_back(false);
		}
		
		// Delete the rotated rectangle
		delete[] robotRect;
	}
	
	return collisions;
}

Point* Environment::rotateRectangle(double _xCenter, double _yCenter, double _angle, double _width, double _height) {
	
	//
	// Calculate the rectangle points
	//
	
	Point centerPoint;
	centerPoint.x = _xCenter;
	centerPoint.y = _yCenter;
	
	// Top-left corner
	Point pTopLeft;
	pTopLeft.x = _xCenter - (_width/2.0);
	pTopLeft.y = _yCenter + (_height/2.0);
	
	// Top-right corner
	Point pTopRight;
	pTopRight.x = _xCenter + (_width/2.0);
	pTopRight.y = _yCenter + (_height/2.0);
	
	// Bottom-left corner
	Point pBotLeft;
	pBotLeft.x = _xCenter - (_width/2.0);
	pBotLeft.y = _yCenter - (_height/2.0);
	
	// Bottom-right corner
	Point pBotRight;
	pBotRight.x = _xCenter + (_width/2.0);
	pBotRight.y = _yCenter - (_height/2.0);
	
	// Transform the points based on the given angle
	pTopLeft = rotatePoint(pTopLeft, centerPoint, _angle);
	pTopRight = rotatePoint(pTopRight, centerPoint, _angle);
	pBotLeft = rotatePoint(pBotLeft, centerPoint, _angle);
	pBotRight = rotatePoint(pBotRight, centerPoint, _angle);
	
	// Create the Rectangle
	Point* robotRect = new Point[4];
	robotRect[0] = pTopLeft;
	robotRect[1] = pTopRight;
	robotRect[2] = pBotRight;
	robotRect[3] = pBotLeft;
	
	return robotRect;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
Position Environment::getStartPosition(bool _adjustment) {

	// Check if the start position needs to be adjusted
	Position currPos = m_robot->getCurrentPosition();
	Position targetPos = getTargetPosition();
	
	Position currPosRot = rotatePosition(currPos, Position(0.0,0.0,0.0), currPos.getTheta());
	Position targetPosRot = rotatePosition(targetPos, Position(0.0,0.0,0.0), targetPos.getTheta());
	
	double xDiff =  targetPosRot.getX() - currPosRot.getX() + START_POSITION_ADJUST_ADDITIONAL_X;
	
	if(_adjustment && xDiff < -0.10) {
	
		// Adjusting
		xDiff = abs(xDiff);
		cout << "Adjusting start position by: " << xDiff << " m" << endl;
		
		double driveSpeed = m_robot->getDriveSpeed();
		double driveTime = xDiff / driveSpeed;
		
		m_robot->setRouteCommand(RouteCommand(driveDirectionBackward, driveTime, 0.0));
		m_robot->stop();
		
		cout << "Adjustment complete !" << endl;
	}
	
	return m_robot->getCurrentPosition();
}

/*
*-------------------------------------------------------------------------------------------------------
*/
Position Environment::getTargetPosition() {

	// Check if we got a parking slot
	if(parkingLines == NULL) {
	
		// No parking slot, use current position
		return m_robot->getCurrentPosition();
	}
	
	// Calculate the middle point of the parking lot
	RobotDimension dim = m_robot->getRobotDimension();
	
	Position targetPos = mp_of_ps(&parkingLines[1], &parkingLines[4], &parkingLines[2], &parkingLines[3]);
	
	double normalY = parkingLines[3].x - parkingLines[2].x;
	double normalX = -(parkingLines[3].y - parkingLines[2].y);
	double normalLength = sqrt(normalX*normalX + normalY*normalY);
	normalY /= normalLength;
	normalX /= normalLength;
	
	double dist = ((targetPos.getX() - parkingLines[2].x) * normalX) + ((targetPos.getY() - parkingLines[2].y) * normalY);
	
	if(dist < ((dim.width+COLLISION_STATE_WIDTH_THRESHOLD)/2)+0.05) {
		targetPos.setX(targetPos.getX() + (((dim.width+COLLISION_STATE_WIDTH_THRESHOLD)/2)+0.05 - dist) * normalX);
		targetPos.setY(targetPos.getY() + (((dim.width+COLLISION_STATE_WIDTH_THRESHOLD)/2)+0.05 - dist) * normalY);
	}

	// Check if the robot is fitting into the parking lot with the current position
	vector<Position> collPos = vector<Position>();
	collPos.push_back(targetPos);
	
	vector<bool> collision = getCollisionState(collPos);
	
	if(collision.at(0)) {
		cout << "Calculated position within the parking lot is too small and still collides !" << endl;
	}
	
	cout << "Parking at: (x/y/theta) " << targetPos.getX() << " / " << targetPos.getY() << " / " << targetPos.getTheta() << endl;
	
	return targetPos;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
Position Environment::rotatePosition(Position _posToRotate, Position _rotateOrigin, double _angle) {

	_rotateOrigin.setX(-1.0*_rotateOrigin.getX());
	_rotateOrigin.setY(-1.0*_rotateOrigin.getY());

	Position transPoint;
	transPoint.setX(_posToRotate.getX() + _rotateOrigin.getX());
	transPoint.setY(_posToRotate.getY() + _rotateOrigin.getY());
	
	double newX = (transPoint.getX() * cos(_angle)) - (transPoint.getY() * sin(_angle));
	double newY = (transPoint.getX() * sin(_angle)) + (transPoint.getY() * cos(_angle));
	
	transPoint.setX(newX);
	transPoint.setY(newY);
	
	_rotateOrigin.setX(-1.0*_rotateOrigin.getX());
	_rotateOrigin.setY(-1.0*_rotateOrigin.getY());

	Position rotatedPoint = Position();
	rotatedPoint.setX(transPoint.getX() + _rotateOrigin.getX());
	rotatedPoint.setY(transPoint.getY() + _rotateOrigin.getY());
	rotatedPoint.setTheta(0.0);
	
	return rotatedPoint;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
Point Environment::rotatePoint(Point _point, Point _rotationOrigin, double _angle) {

	_rotationOrigin.x = -1.0*_rotationOrigin.x;
	_rotationOrigin.y = -1.0*_rotationOrigin.y;
	
	Point transPoint;
	transPoint.x = _point.x + _rotationOrigin.x;
	transPoint.y = _point.y + _rotationOrigin.y;
	
	double newX = (transPoint.x * cos(_angle)) - (transPoint.y * sin(_angle));
	double newY = (transPoint.x * sin(_angle)) + (transPoint.y * cos(_angle));
	
	transPoint.x = newX;
	transPoint.y = newY;
	
	_rotationOrigin.x = -1.0*_rotationOrigin.x;
	_rotationOrigin.y = -1.0*_rotationOrigin.y;
	
	Point rotatedPoint;
	rotatedPoint.x = transPoint.x + _rotationOrigin.x;
	rotatedPoint.y = transPoint.y + _rotationOrigin.y;
	
	return rotatedPoint;
}

