/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Master of Science - Computergraphics and Image Processing
 *
 * Robotics - Prof. Dr. Max Fischer
 * "Autonomous self-parking robot"
 * 
 * System: Linux, C++
 */

/**
 * USABILITY:
 *  Copyright (c) 2006 
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * Class Description:
 *	Environment taking care of transforming the laser scan data into a useable format
 *	and of the collision detection
 *
 * @author Thomas Wöllert
 */

#ifndef ENVIRONMENT_H
#define ENVIRONMENT_H

#include <vector>
#include "Position.h"
#include "Robot.h"
#include "cv.h"
#include "collision_test.h"

using namespace std;

class Environment {

	protected:
		// The robot to use
		Robot* m_robot;
		
		// The max width/height/angle of the environment
		double m_areaWidth;
		double m_areaHeight;
		double m_areaAngle;

		// The parking slot
		Point* parkingLines;
	
	public:
		/**
		* Constructor
		* @param Robot* the robot to use
		* @param double the max width of the environment
		* @param double the max height of the environment
		* @param double the max angle of the environment in radian (normally dtor(360.0))
		*/
		Environment(Robot* _robot, double _width, double _height, double _angle);
	
		/**
		* Destructor
		*/
		virtual ~Environment();
		
		/**
		* Retrieve the width of the area
		* @return the width
		*/
		double getAreaWidth();
		
		/**
		* Retrieve the height of the area
		* @return the height
		*/
		double getAreaHeight();
		
		/**
		* Retrieve the angle of the area
		* @return the angle
		*/
		double getAreaAngle();
		Point* getParkingLines();
		void setParkingLines(Point leftBorderPoint, Point leftTopPoint, Point leftBottomPoint, Point rightBottomPoint, Point rightTopPoint, Point rightBorderPoint);

		/**
		* Retrieve the collision state for a given list of positions
		* @param vector<Position>& the positions to test
		* @return vector of booleans (true = obstacle, false = free)
		*/
		vector<bool> getCollisionState(const vector<Position>& _posList);
		
		/**
		* Get the start position to use
		* @param bool with (true) or without (true) adjustment
		* @return the start position
		*/
		Position getStartPosition(bool _adjustment);
		
		/**
		* Get the target position to use
		* @return the target position
		*/
		Position getTargetPosition();
		
	protected:
		/**
		* Rotate the given position around a given origin by a given angle
		* @param Position the position to rotate
		* @param Position the rotation origin
		* @param double the angle
		* @return the new rotated position
		*/
		Position rotatePosition(Position _posToRotate, Position _rotateOrigin, double _angle);
		
		Point* rotateRectangle(double _xCenter, double _yCenter, double _angle, double _width, double _height);
		Point rotatePoint(Point _point, Point _rotationOrigin, double _angle);
};

#endif // ENVIRONMENT_H
