#include "Robot.h"
#include "CVCoordDrawingArea.h"

#include <iostream>
#include <math.h>
#include <time.h>

extern CVCoordDrawingArea* g_CoordDrawingArea;

/*
*-------------------------------------------------------------------------------------------------------
*/
Robot::Robot(RobotDimension _dim, double _driveSpeed, const string _hostname, unsigned int _port) {

	dim = _dim;

	playerClient = new PlayerClient(_hostname, _port);
	pProxy = new Position2dProxy(playerClient, 0);
	lProxy = new LaserProxy(playerClient, 0);
	
	cout << "Before Config" << endl;
	#ifndef SIMULATION
		//lProxy->Configure(0.0, dtor(180.0), 50, 100, false);
	#endif 
	cout << "After Config" << endl;
	//sProxy = new SonarProxy(playerClient, 0);
	//gProxy = new Graphics2dProxy(playerClient, 0);

	driveSpeed = _driveSpeed;
}

Robot::~Robot() {

	if(pProxy)
		delete pProxy;

	if(lProxy != NULL) {
		delete lProxy;
	}
	
	if(sProxy != NULL) {
		delete sProxy;
	}

	if(gProxy != NULL) {
		delete gProxy;
	}

	if(playerClient != NULL) {
		delete playerClient;
	}
}

/*
*-------------------------------------------------------------------------------------------------------
*/
RobotDimension Robot::getRobotDimension() {
	return dim;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
double Robot::getDriveSpeed() {
	return driveSpeed;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void Robot::updateSensorData() {
	playerClient->Read();
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void Robot::setMotorEnable(bool _enable) {
	pProxy->SetMotorEnable(_enable);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
Position Robot::getCurrentPosition() {
	
	// Simulation Yaw: PI->0 0->-PI
	// Real Yaw: 0->2*PI
	double yaw = pProxy->GetYaw();
	
	#ifdef SIMULATION
		// Convert Simulation yaw to real yaw
	 	if(yaw < 0) {
	 		yaw += 2*M_PI;
	 	}
	#else
		// Nothing to do in that case, already getting the correct yaw
	#endif

	return Position(pProxy->GetXPos(), pProxy->GetYPos(), yaw);
}

double Robot::getCurrentSpeed() {
	double x = pProxy->GetXSpeed();
	double y = pProxy->GetYSpeed();
	return sqrt(x * x + y * y);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
unsigned int Robot::getLaserProxyCount() {
	return lProxy->GetCount();
}

/*
*-------------------------------------------------------------------------------------------------------
*/
double* Robot::getLaserProxyData() {
	double* lProxyData = new double[this->getLaserProxyCount()];

	for(unsigned int i = 0; i < this->getLaserProxyCount(); i++) {
		lProxyData[i] = (*lProxy)[i];
	}

	return lProxyData;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
unsigned int Robot::getSonarProxyCount() {
	return sProxy->GetCount();
}

/*
*-------------------------------------------------------------------------------------------------------
*/
double* Robot::getSonarProxyData() {
	double* sProxyData = new double[this->getSonarProxyCount()];

	for(unsigned int i = 0; i < this->getSonarProxyCount(); i++) {
		sProxyData[i] = (*sProxy)[i];
	}

	return sProxyData;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void Robot::move(double _driveSpeed, double _turnRate) {
	pProxy->SetSpeed(_driveSpeed, _turnRate);
}

void Robot::move_to(Position& p) {
	pProxy->GoTo(p.getX(), p.getY(), p.getTheta());
}

void Robot::move_to(Position p) {
	pProxy->GoTo(p.getX(), p.getY(), p.getTheta());
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void Robot::stop() {
	pProxy->SetSpeed(0.0, 0.0);
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void Robot::setListOfRouteCommands( const list<RouteCommand>& _cmdList ) {

	// Run through the given list of route commands
	list<RouteCommand>::const_iterator iter;

	for(iter = _cmdList.begin(); iter != _cmdList.end(); iter++) {

		g_CoordDrawingArea->drawRobot(getCurrentPosition(), getRobotDimension(), 1, g_CoordDrawingArea->createColor("red"));
		g_CoordDrawingArea->updateWindow();

		// Set the current command
		setRouteCommand(*iter);

		//stop();
	}

	// All commands done, stop movement
	this->stop();	
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void Robot::setRouteCommand(const RouteCommand _cmd) {

	// Determine the needed drive speed (forward/backward) by checking the route command
	double currentDriveSpeed = 0.0;

	switch(_cmd.getDrive()) {

		case driveDirectionForward:
			currentDriveSpeed = driveSpeed;
		break;

		case driveDirectionBackward:
			currentDriveSpeed = -1.0*driveSpeed;
		break;
	}

	// Calculate the needed turn rate in degrees/s based on the given deltaTheta and the time
	double turnRate = _cmd.getDeltaTheta() / _cmd.getTime();
	
	g_CoordDrawingArea->drawRobot(getCurrentPosition(), getRobotDimension(), 1, g_CoordDrawingArea->createColor("red"));
	g_CoordDrawingArea->updateWindow();

	// Move the robot
	pProxy->SetSpeed(currentDriveSpeed, turnRate);
	
	// Sleep till route command is complete
	usleep((int)(_cmd.getTime()*1000000));
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void Robot::driveCarLikeTo(Position _target, RouteCommand _cmd, double _xThreshold, double _yThreshold, double _thetaThreshold) {

	Position currentPos = getCurrentPosition();
	
	// Determine the needed drive speed (forward/backward) by checking the route command
	double currentDriveSpeed = 0.0;

	switch(_cmd.getDrive()) {

		case driveDirectionForward:
			currentDriveSpeed = driveSpeed;
		break;

		case driveDirectionBackward:
			currentDriveSpeed = -1.0*driveSpeed;
		break;
	}
	
	// Calculate the needed turn rate in degrees/s based on the given deltaTheta and the time
	double turnRate = _cmd.getDeltaTheta() / _cmd.getTime();
	
	g_CoordDrawingArea->drawRobot(getCurrentPosition(), getRobotDimension(), 1, g_CoordDrawingArea->createColor("red"));
	g_CoordDrawingArea->updateWindow();
	
	// Move the robot, till it reached the target (taking care of the given threshold)
	//struct timespec startTime;
	//clock_gettime(CLOCK_REALTIME, &startTime);
	struct timeval startTime;
	gettimeofday( &startTime , NULL);
	
	while(true) {
	
		// 1st Criteria - Check if we have reached the target
		currentPos = getCurrentPosition();
		
		if(currentPos.getX() <= (_target.getX()+_xThreshold) && currentPos.getX() >= (_target.getX()-_xThreshold)) {
			if(currentPos.getY() <= (_target.getY()+_yThreshold) && currentPos.getY() >= (_target.getY()-_yThreshold)) {
				if(currentPos.getTheta() <= (_target.getTheta()+_thetaThreshold) && currentPos.getTheta() >= (_target.getTheta()-_thetaThreshold)) {
				
					// Target reached
					break;
				}		
			}	
		}
		
		// 2nd Criteria - Abort if the command duration time has been reached
//		struct timespec currentTime;
//		clock_gettime(CLOCK_REALTIME, &currentTime);
		struct timeval currentTime;
		gettimeofday( &currentTime , NULL);
		
		//double duration = ((double)(currentTime.tv_sec - startTime.tv_sec)) + ((double)(currentTime.tv_nsec - startTime.tv_nsec))/((double)1000000000 );
		
		double duration = ((double)(currentTime.tv_sec - startTime.tv_sec)) + ((double)(currentTime.tv_usec - startTime.tv_usec))/((double)1000000 );
		
		if(duration >= _cmd.getTime()) {
		
			// Max time reached
			break;
		}
	
		pProxy->SetSpeed(currentDriveSpeed, turnRate);
		usleep(100);
	}
}

/*
*-------------------------------------------------------------------------------------------------------
*/
PlayerClient* Robot::getPlayerClient() {
	return playerClient;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
Position2dProxy* Robot::getPosition2dProxy() {
	return pProxy;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
LaserProxy* Robot::getLaserProxy() {
	return lProxy;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
SonarProxy* Robot::getSonarProxy() {
	return sProxy;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
Graphics2dProxy* Robot::getGraphicsProxy() {
	return gProxy;
}
