/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Master of Science - Computergraphics and Image Processing
 *
 * Robotics - Prof. Dr. Max Fischer
 * "Autonomous self-parking robot"
 * 
 * System: Linux, C++
 */

/**
 * USABILITY:
 *  Copyright (c) 2006 
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * Class Description:
 *	This class represents the robot, its sensors and movement
 *
 * Authors: 
 *	@author Thomas Wöllert, Beatrix Müller, Herbert Hundsnurscher
 */

#ifndef ROBOT_H
#define ROBOT_H

#include <string>
#include <libplayerc++/playerc++.h>
#include <list>

#include "Position.h"
#include "RouteCommand.h"

using namespace std;
using namespace PlayerCc;

/**
* Representing the robot size
*/
struct RobotDimension {

	// The length of the robot
	double length;
	
	// The width of the robot
	double width;
	
	// The x/y-Offset of the rotation center
	double xOffset;
	double yOffset;
};

class Robot {

	protected:
		// The player client representing this robot
		PlayerClient* playerClient;

		// The position proxy responsible for movement and position infos
		Position2dProxy* pProxy;

		// The laser proxy representing the SICK laser scanner
		LaserProxy* lProxy;

		// The laser proxy representing the Sonar sensors
		SonarProxy* sProxy;

		// The graphics proxy representing the graphical interface adjacent to this robot
		Graphics2dProxy* gProxy;
	
		// The given drive speed of this robot
		double driveSpeed;

		// Robot dimensions
		RobotDimension dim;

	public:
		/**
		* Robot Constructor
		* @param RobotDimension the robot dimensions
		* @param double the drive speed (m/s)
		* @param string the hostname of the robot to connect to
		* @param uint the port of the robot to connect to
		*/
		Robot(RobotDimension _dim, double _driveSpeed, const string _hostname, unsigned int _port);
	
		/**
		* Destructor
		*/
		virtual ~Robot();

		/**
		* Retrieve the robot dimensions
		* @return the dimensions
		*/
		RobotDimension getRobotDimension();

		/**
		* Retrieve the drive speed (m/s)
		* @return the drive speed
		*/
		double getDriveSpeed();
	
		/**
		* Update all sensor data of the robot (position, laser data, etc.)
		*/
		void updateSensorData();

		/**
		* Enable/Disable the motor
		* @param bool true to enable, false to disable
		*/
		void setMotorEnable(bool _enable);

		/**
		* Retrieve the current position of the robot
		* @return Position
		*/
		Position getCurrentPosition();
		double getCurrentSpeed();
	
		/**
		* Set a list of route commands the robot is following
		* (returns when all route commands are done)
		* @param list<RouteCommand>& the route command list
		*/
		void setListOfRouteCommands(const list<RouteCommand>& _cmdList);

		/**
		* Set a single route command the robot is following
		* (returns when the given route command is done)
		* @param RouteCommand the route command
		*/
		void setRouteCommand(const RouteCommand _cmd);

		/**
		* Manual by hand movement (returns immediatly, drives till stop() is called)
		* @param double drive speed in meters per second
		* @param turnRate turn rate in degrees per second
		*/
		void move(double _driveSpeed, double _turnRate);
		void move_to(Position& p);
		void move_to(Position p);
		
		void driveCarLikeTo(Position _target, RouteCommand _cmd, double _xThreshold, double _yThreshold, double _thetaThreshold);

		/**
		* Stops all movement of the robot
		*/
		void stop();

		/**
		* Retrieve the laser data count
		* @return uint the number of information in the laser data array
		*/
		unsigned int getLaserProxyCount();
		
		/**
		* Retrieve the laser data array (distances)
		* @return double* the array of laser data
		*/
		double* getLaserProxyData();

		/**
		* Retrieve the sonar data count
		* @return uint the number of information in the sonar data array
		*/
		unsigned int getSonarProxyCount();
		
		/**
		* Retrieve the sonar data array (distances)
		* @return double* the array of sonar data
		*/
		double* getSonarProxyData();

		/**
		* Retrieve the adjacent player client
		* @return PlayerClient* the player client
		*/
		PlayerClient* getPlayerClient();

		/**
		* Retrieve the adjacent position2d proxy
		* @return Position2dProxy* the position2d proxy
		*/
		Position2dProxy* getPosition2dProxy();

		/**
		* Retrieve the adjacent laser proxy
		* @return LaserProxy* the laser proxy
		*/
		LaserProxy* getLaserProxy();

		/**
		* Retrieve the adjacent sonar proxy
		* @return SonarProxy* the sonar proxy
		*/
		SonarProxy* getSonarProxy();
	
		/**
		* Retrieve the adjacent graphics proxy to paint into it
		* @return Graphics2dProxy* the graphics proxy
		*/
		Graphics2dProxy* getGraphicsProxy();
};

#endif // ROBOT_H
