#include "UserPlaner.h"
#include "Car.h"
#include "CVCoordDrawingArea.h"
#include "Environment.h"
#include "AStarPlaner.h"

// Headerfile Umgebungsmodell
#include "FindParkingLot.h"
#include "Point.h"

#include "cv.h"
#include "highgui.h"

#include <iostream>

extern Robot* g_Robot;
extern CVCoordDrawingArea* g_CoordDrawingArea;
extern Environment* g_Environment;
extern const double maxSteeringAngle;

/*
*-------------------------------------------------------------------------------------------------------
*/
UserPlaner::UserPlaner(Car* _car, double _deltaT) {
	car = _car;
	deltaT = _deltaT;
}

UserPlaner::~UserPlaner() {}

/*
*-------------------------------------------------------------------------------------------------------
*/
void UserPlaner::run() {

	// Coordinates from the environment
	struct Point leftBorderPoint;
	struct Point leftTopPoint;
	struct Point leftBottomPoint;
	struct Point rightBottomPoint;
	struct Point rightTopPoint;
	struct Point rightBorderPoint;

	while(true) {

		// Give the user manual control and wait for the parking key
		int key = waitForParkingKey();
	
		// Check if the user aborted or wanted to initialize the parking
		if(key == KEY_START_PARKING) {
		
			cout << "Initializing parking ..." << endl;
			
			// Stop the car first
			car->stop();

			// Read the laser data and create the environment model
			Position currentPos = g_Robot->getCurrentPosition();
			
			#ifdef INTERPOLATE_LASER_DATA
			// Raise the number of laser data samples from 181 to 362 by linear interpolation
			laserdata extData = interpolateLaserData(g_Robot->getLaserProxyData(), g_Robot->getLaserProxyCount());
			
			if(findParkingLot(extData.data,extData.count,
				&currentPos,leftBorderPoint,leftTopPoint,leftBottomPoint,
				rightBottomPoint,rightTopPoint,rightBorderPoint) )  {
			#else
			if(findParkingLot(g_Robot->getLaserProxyData(),g_Robot->getLaserProxyCount(),
				&currentPos,leftBorderPoint,leftTopPoint,leftBottomPoint,
				rightBottomPoint,rightTopPoint,rightBorderPoint) )  {
			#endif
				
				cout << "Parking lot found ..." << endl;
				
				// Set the parking lines
				g_Environment->setParkingLines(leftBorderPoint,leftTopPoint,leftBottomPoint,rightBottomPoint,rightTopPoint,rightBorderPoint);
				
				// Start parking
				initParking();
			} else {
				cout << "No Parking Lot found !!!" << endl;
			}
		} else if(key == KEY_DRIVE_REVERSE) {
		
			// drive from parking lot to start position
			if (starPlaner != NULL) {
				starPlaner->driveReverse();
			}			
			
		}
	}
}

/*
*-------------------------------------------------------------------------------------------------------
*/
laserdata UserPlaner::interpolateLaserData(double* _dataArray, int _dataCount) {
	
	// Create the new extended array
	int extDataCount = (_dataCount)*2;
	double* extDataArray = new double[extDataCount];
	
	// Copy the original data onto the proper spots in the new array
	for(int i = 0; i < _dataCount; i++) {
	
		extDataArray[i*2] = _dataArray[i];
	}
	
	// Run through the larger copied array and create the data points in-between
	for(int i = 1; i < extDataCount; i = i+2) {
	
		double diff = abs(extDataArray[i+1] - extDataArray[i-1]);
		
		if(extDataArray[i-1] < extDataArray[i+1]) {
			extDataArray[i] = extDataArray[i-1] + (diff/2);
		} else {
			extDataArray[i] = extDataArray[i+1] + (diff/2);
		}
	}
	
	laserdata finaldata;
	finaldata.count = extDataCount;
	finaldata.data = extDataArray;
	
	return finaldata;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
int UserPlaner::waitForParkingKey() {

	cout << "Waiting for user input ..." << endl;

	Position oldPos(0.0, 0.0, 0.0);
	
	// The pressed key
	int key = -1;

	// Go into the loop
	while(key != KEY_START_PARKING && key != KEY_DRIVE_REVERSE) {
	
		Position currentPos = g_Robot->getCurrentPosition();
		
		if(currentPos.getX() != oldPos.getX() || currentPos.getY() != oldPos.getY() || currentPos.getTheta() != oldPos.getTheta()) {
			g_CoordDrawingArea->drawRobot(currentPos, g_Robot->getRobotDimension(), 1, g_CoordDrawingArea->createColor("red"));
			g_CoordDrawingArea->updateWindow();
			oldPos = currentPos;
		}
		
		// Check which key has been pressed
		switch(key) {
			
			case KEY_MOVE_FORWARD_STRAIGHT:
				// Move forward
				car->drive(driveDirectionForward, driveSteeringStraight);
			break;
			
			case KEY_MOVE_FORWARD_LEFT:
				// Move forward to the left
				car->drive(driveDirectionForward, driveSteeringLeft);
			break;
			
			case KEY_MOVE_FORWARD_RIGHT:
				// Move forward to the right
				car->drive(driveDirectionForward, driveSteeringRight);
			break;
			
			case KEY_MOVE_BACKWARD_STRAIGHT:
				// Move backward
				car->drive(driveDirectionBackward, driveSteeringStraight);
			break;
			
			case KEY_MOVE_BACKWARD_LEFT:
				// Move backward to the left
				car->drive(driveDirectionBackward, driveSteeringLeft);
			break;
			
			case KEY_MOVE_BACKWARD_RIGHT:
				// Move backward to the right
				car->drive(driveDirectionBackward, driveSteeringRight);
			break;
			
			case KEY_MOVE_STOP:
				car->stop();
			break;
			
			default:
				// Stop any movement
				//cout << "stop" << endl;
				// car->stop();
				;
		}
		
		// Get the new key
		key = cvWaitKey(10);

		if(key == -1 && js.joystickSupported()) {
		
			// get joystick data
			bool had_events = js.wait_event(100);
		
			// Buttons
			if(js.had_button_event) {
			
				if(js.pressedButton == 1) {
				
					cout << "Button: START PARKING" << endl;
					key = KEY_START_PARKING;
				} else if(js.pressedButton == 3) {
				
					cout << "Button: DRIVE REVERSE" << endl;
					key = KEY_DRIVE_REVERSE;
				}
			}
			
			if(had_events) {
			
				// printf("js events: %.2f %.2f ", js.throttle, js.steering);
				double speed = -js.throttle * .25; // todo: max 1.5m/s speed
				double steering_angle = -js.steering * maxSteeringAngle;
				static double last_steering_angle = steering_angle;
				double cs = car->m_robot->getCurrentSpeed();
				Position cp = car->m_robot->getCurrentPosition();
				// cout << "current position: " << cp << " speed:" << cs << " \n";
				double delta_t = 2;
				Position np = car->getNextPosition(delta_t, cp, cs, last_steering_angle, speed, steering_angle);
				double distance = np - cp;
				// #printf("joystick command has distance of %.3f\n", distance);

				double vr = (np.getTheta() - cp.getTheta()) / delta_t;
				double sign = js.throttle < 0 ? -1.0 : 1.0;
				double vx = sign * distance / delta_t;

				vx = vx * (1 - 0.4 * fabs (vr));

				if(fabs(vx) < 0.1) {
					vx = 0;
					vr = 0;
				}
					
				car->m_robot->move(-vx, vr);

				last_steering_angle = steering_angle;
			}
		}
	}
	
	return key;
}

/*
*-------------------------------------------------------------------------------------------------------
*/
void UserPlaner::initParking() {

	if (starPlaner != NULL) {
		cout << "Deleting" << endl;
		delete starPlaner;
		cout << "Delete done" << endl;
	}
	
	starPlaner = new AStarPlaner(car, deltaT);
	starPlaner->run();
}
