/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * Master of Science - Computergraphics and Image Processing
 *
 * Robotics - Prof. Dr. Max Fischer
 * "Autonomous self-parking robot"
 * 
 * System: Linux, C++
 */

/**
 * USABILITY:
 *  Copyright (c) 2006 
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * Class Description:
 *	Main class of the application
 *
 * Authors: 
 *	Thomas Wöllert, Beatrix Müller, Herbert Hundsnurscher
 */

#include <stdio.h>
#include <math.h>
#include <iostream>
#include <pthread.h>

#include "Robot.h"
#include "RouteCommand.h"
#include "Car.h"
#include "UserPlaner.h"
#include "AStarPlaner.h"
#include "CVCoordDrawingArea.h"
#include "Environment.h"

extern Robot* g_Robot;
extern CVCoordDrawingArea* g_CoordDrawingArea;
extern Environment* g_Environment;

#include "cv.h"
#include "highgui.h"

// SWITCH FOR MULTITHREADED PLAYER
// When running the multithreaded player the internal thread of the player client is started
// which is sufficient to keep updating the sensor data for other parts of the program.
// When running the non-multithreaded player the update needs to be taken care of by our local POSIX
// thread, but in this case the thread of the player cannot be started.
//
// Currently it works better when doing the update ourself !
//
//#define _MULTITHREADED_PLAYER_

/**
* Some constants
*/
const static double driveSpeed = 0.1;
double maxSteeringAngle = dtor(22.0);

#ifdef SIMULATION
	const static string robotHostname = "localhost";
#else
	const static string robotHostname = "10.28.7.50";
#endif

const static unsigned int robotPort = 6665;
const static double deltaT = 0.6;
static bool initRobotDone = false;

#ifdef SIMULATION
	const static double robotLength = 0.44;
	const static double robotWidth = 0.33;
	const static double robotXOffset = -0.04;
	const static double robotYOffset = 0.0;
#else
	const static double robotLength = 0.49;
	const static double robotWidth = 0.38;
	const static double robotXOffset = -0.04;
	const static double robotYOffset = 0.0;
#endif

const static double areaWidth = 1.5;
const static double areaHeight = 3.0;
const static double areaAngle = dtor(360.0);

/**
* Initializes and shows the GUI (window and drawing area)
*/
void initGUI() {

	// GUI Init
	cout << "Initializing gui ..." << endl;

	// Create the drawing area and paint the grid
	g_CoordDrawingArea = new CVCoordDrawingArea(-3.0, 3.0, -3.0, 3.0, 640, 480);
	g_CoordDrawingArea->showWindow("Autonomous self-parking robot", CV_WINDOW_AUTOSIZE);
	g_CoordDrawingArea->setBackgroundColor(g_CoordDrawingArea->createColor("white"));
	g_CoordDrawingArea->drawGrid(1.0, 1.0, 1, g_CoordDrawingArea->createColor("lightgray"));
	g_CoordDrawingArea->drawGridText(1.0, 1.0, g_CoordDrawingArea->createColor("lightgray"));
	g_CoordDrawingArea->updateWindow();

	cout << "Initialization of gui done" << endl;

	return;
}

/**
* Robot Run-Method for the Posix Thread
*/
void* runRobot(void*) {

	try {
		cout << "Initializing robot & environment ..." << endl;
	
		//
		// INITIALIZATION
		//		
	
		// Create the robot and loose the breaks
		RobotDimension dim;
		dim.length = robotLength;
		dim.width = robotWidth;
		dim.xOffset = robotXOffset;
		dim.yOffset = robotYOffset;

		g_Robot = new Robot(dim, driveSpeed, robotHostname, robotPort);
		g_Robot->setMotorEnable(true);

		// Create the environment
		g_Environment = new Environment(g_Robot, areaWidth, areaHeight, areaAngle);
	
		#ifdef _MULTITHREADED_PLAYER_
			// Starting the player client thread (when using the multithreaded player)
			g_Robot->getPlayerClient()->StartThread();
		#endif

		#ifndef _MULTITHREADED_PLAYER_
			// Make a manual data update (when using the non-multithreaded player)
			g_Robot->updateSensorData();
		#endif

		Position startPos = g_Robot->getCurrentPosition();
		cout << "Initialization of robot done" << endl;
		cout << "Current robot position (x/y/theta): " << startPos.getX() << " / " << startPos.getY() << " / " << startPos.getTheta() << endl;
	
		//
		// KEEP RUNNING TO UPDATE THE SENSOR DATA (not needed when running the Multithreaded Player)
		//

		initRobotDone = true;
	
		#ifndef _MULTITHREADED_PLAYER_
			for(;;) {
				g_Robot->updateSensorData();
				usleep(100);
			}
		#endif
	}
	catch (PlayerError e)
	{
		std::cerr << e << std::endl;
		return NULL;
	}

	return NULL;
}

/**
* Planer Run-Method for the Posix Thread
*/
void runPlaner() {

	cout << "Initializing planer ..." << endl;

	//
	// INITIALIZATION
	//

	// Create the car model wrapping the robot (Robot*, px, nx, py, ny, wheelbase)
	//Car car = Car(g_Robot, maxSteeringAngle, 0.26, -0.18, 16.5, -16.5, 0.44);
	Car car = Car(g_Robot, maxSteeringAngle, 0.26, -0.18, 0.17, -0.17, 0.44);

	// Create the external control planer (for the user to control the robot)
	UserPlaner userPlaner = UserPlaner(&car, deltaT);

	//
	// RUN
	//
	
	userPlaner.run();

	//
	// AFTERMATH
	//

	// Stop all movement
	g_Robot->stop();
	
	return;
}

/**
* Main Method
* @param int number of input arguments (optional)
* @param char* the given input arguments (optional)
*/
int main(int argc, char* argv[]) {

	// The Robot thread and its attributes
	pthread_t threadRobot;
	pthread_attr_t threadAttrRobot;
	pthread_attr_init(&threadAttrRobot);

	// Create and show the GUI
	initGUI();

	// Create and run the Robot thread
	int ret = pthread_create(&threadRobot, &threadAttrRobot, &runRobot, NULL);
	if(ret != 0) {
		cout << "pthread_create() for Robot failed !" << endl;
		return ret;
	}

	while(!initRobotDone) {
		// Wait
		usleep(500);
	}

	// Run the planer
	runPlaner();

	return 0;
}
